import fnmatch
import json
import re
import sys
import time
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Tuple

import boto3
import pytz
import sentry_sdk
from asgiref.sync import sync_to_async
from botocore.exceptions import ClientError, ParamValidationError
from cloudaux import CloudAux
from cloudaux.aws.decorators import rate_limited
from cloudaux.aws.iam import get_managed_policy_document, get_policy
from cloudaux.aws.s3 import (
    get_bucket_location,
    get_bucket_policy,
    get_bucket_resource,
    get_bucket_tagging,
)
from cloudaux.aws.sns import get_topic_attributes
from cloudaux.aws.sqs import get_queue_attributes, get_queue_url, list_queue_tags
from cloudaux.aws.sts import boto3_cached_conn
from dateutil.parser import parse
from deepdiff import DeepDiff
from parliament import analyze_policy_string, enhance_finding
from policy_sentry.util.arns import get_account_from_arn, parse_arn

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    BackgroundCheckNotPassedException,
    InvalidInvocationArgument,
    MissingConfigurationValue,
)
from consoleme.lib.account_indexers.aws_organizations import (
    retrieve_org_structure,
    retrieve_scps_for_organization,
)
from consoleme.lib.aws_config.aws_config import query
from consoleme.lib.cache import (
    retrieve_json_data_from_redis_or_s3,
    store_json_results_in_redis_and_s3,
)
from consoleme.lib.generic import sort_dict
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.redis import RedisHandler, redis_hget, redis_hgetex, redis_hsetex
from consoleme.models import (
    CloneRoleRequestModel,
    RoleCreationRequestModel,
    ServiceControlPolicyArrayModel,
    ServiceControlPolicyModel,
)

ALL_IAM_MANAGED_POLICIES: dict = {}
ALL_IAM_MANAGED_POLICIES_LAST_UPDATE: int = 0

log = config.get_logger(__name__)
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
red = RedisHandler().redis_sync()


@rate_limited()
def create_managed_policy(cloudaux, name, path, policy, description):
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "cloudaux": cloudaux,
        "name": name,
        "path": path,
        "policy": policy,
        "description": "description",
        "message": "Creating Managed Policy",
    }
    log.debug(log_data)

    cloudaux.call(
        "iam.client.create_policy",
        PolicyName=name,
        Path=path,
        PolicyDocument=json.dumps(policy, indent=2),
        Description=description,
    )


async def needs_updating(existing_policy, new_policy):
    diff = DeepDiff(
        existing_policy, new_policy, ignore_order=True, report_repetition=True
    )
    if diff:
        return True
    return False


async def update_managed_policy(cloudaux, policy_name, new_policy, policy_arn):
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "new_policy": new_policy,
        "policy_name": policy_name,
        "policy_arn": policy_arn,
        "message": "Updating managed policy",
    }
    log.debug(log_data)

    current_policy_versions = []
    default_policy_index = 0
    versions = await sync_to_async(cloudaux.call)(
        "iam.client.list_policy_versions", PolicyArn=policy_arn
    )
    oldest_policy_version = -1
    oldest_timestamp = None
    for i, version in enumerate(versions.get("Versions", [])):
        if version["IsDefaultVersion"]:
            default_policy_index = i
        current_policy_versions.append(version)
        if oldest_policy_version == -1 or oldest_timestamp > version["CreateDate"]:
            oldest_policy_version = i
            oldest_timestamp = version["CreateDate"]

    if len(current_policy_versions) == 5:
        pop_position = oldest_policy_version
        # Want to make sure we don't pop the default version so arbitrarily set position to oldest + 1 mod N
        # if default is also the oldest
        if default_policy_index == oldest_policy_version:
            pop_position = (oldest_policy_version + 1) % len(current_policy_versions)
        await sync_to_async(cloudaux.call)(
            "iam.client.delete_policy_version",
            PolicyArn=policy_arn,
            VersionId=current_policy_versions.pop(pop_position)["VersionId"],
        )

    await sync_to_async(cloudaux.call)(
        "iam.client.create_policy_version",
        PolicyArn=policy_arn,
        PolicyDocument=json.dumps(new_policy, indent=2),
        SetAsDefault=True,
    )


async def create_or_update_managed_policy(
    new_policy,
    policy_name,
    policy_arn,
    description,
    policy_path="/",
    existing_policy=None,
    **conn_details,
):
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "new_policy": new_policy,
        "policy_name": policy_name,
        "policy_arn": policy_arn,
        "description": description,
        "policy_path": policy_path,
        "existing_policy": existing_policy,
        "conn_details": conn_details,
    }

    ca = await sync_to_async(CloudAux)(**conn_details)

    if not existing_policy:
        log_data["message"] = "Policy does not exist. Creating"
        log.debug(log_data)
        await sync_to_async(create_managed_policy)(
            ca, policy_name, policy_path, new_policy, description
        )
        return

    log_data["message"] = "Policy exists and needs to be updated"
    log.debug(log_data)
    # Update the managed policy
    await update_managed_policy(ca, policy_name, new_policy, policy_arn)


async def get_all_iam_managed_policies_for_account(account_id):
    global ALL_IAM_MANAGED_POLICIES_LAST_UPDATE
    global ALL_IAM_MANAGED_POLICIES

    policy_key: str = config.get(
        "redis.iam_managed_policies_key", "IAM_MANAGED_POLICIES"
    )
    current_time = time.time()
    if current_time - ALL_IAM_MANAGED_POLICIES_LAST_UPDATE > 500:
        red = await RedisHandler().redis()
        ALL_IAM_MANAGED_POLICIES = await sync_to_async(red.hgetall)(policy_key)
        ALL_IAM_MANAGED_POLICIES_LAST_UPDATE = current_time

    if ALL_IAM_MANAGED_POLICIES:
        return json.loads(ALL_IAM_MANAGED_POLICIES.get(account_id, "[]"))
    else:
        s3_bucket = config.get("account_resource_cache.s3.bucket")
        s3_key = config.get(
            "account_resource_cache.s3.file",
            "account_resource_cache/cache_{resource_type}_{account_id}_v1.json.gz",
        ).format(resource_type="managed_policies", account_id=account_id)
        return await retrieve_json_data_from_redis_or_s3(
            s3_bucket=s3_bucket, s3_key=s3_key, default=[]
        )


async def get_resource_account(arn: str) -> str:
    """Return the AWS account ID that owns a resource.

    In most cases, this will pull the ID directly from the ARN.
    If we are unsuccessful in pulling the account from ARN, we try to grab it from our resources cache
    """
    red = await RedisHandler().redis()
    resource_account: str = get_account_from_arn(arn)
    if resource_account:
        return resource_account

    resources_from_aws_config_redis_key: str = config.get(
        "aws_config_cache.redis_key", "AWSCONFIG_RESOURCE_CACHE"
    )

    if not red.exists(resources_from_aws_config_redis_key):
        # This will force a refresh of our redis cache if the data exists in S3
        await retrieve_json_data_from_redis_or_s3(
            redis_key=resources_from_aws_config_redis_key,
            s3_bucket=config.get("aws_config_cache_combined.s3.bucket"),
            s3_key=config.get(
                "aws_config_cache_combined.s3.file",
                "aws_config_cache_combined/aws_config_resource_cache_combined_v1.json.gz",
            ),
            redis_data_type="hash",
        )

    resource_info = await redis_hget(resources_from_aws_config_redis_key, arn)
    if resource_info:
        return json.loads(resource_info).get("accountId", "")
    elif "arn:aws:s3:::" in arn:
        # Try to retrieve S3 bucket information from S3 cache. This is inefficient and we should ideally have
        # retrieved this info from our AWS Config cache, but we've encountered problems with AWS Config historically
        # that have necessitated this code.
        s3_cache = await retrieve_json_data_from_redis_or_s3(
            redis_key=config.get("redis.s3_buckets_key", "S3_BUCKETS"),
            redis_data_type="hash",
        )
        search_bucket_name = arn.split(":")[-1]
        for bucket_account_id, buckets in s3_cache.items():
            buckets_j = json.loads(buckets)
            if search_bucket_name in buckets_j:
                return bucket_account_id
    return ""


async def get_resource_policy(account: str, resource_type: str, name: str, region: str):
    try:
        details = await fetch_resource_details(account, resource_type, name, region)
    except ClientError:
        # We don't have access to this resource, so we can't get the policy.
        details = {}

    # Default policy
    default_policy = {"Version": "2012-10-17", "Statement": []}

    # When NoSuchBucketPolicy, the above method returns {"Policy": {}}, so we default to blank policy
    if "Policy" in details and "Statement" not in details["Policy"]:
        details = {"Policy": default_policy}

    # Default to a blank policy
    return details.get("Policy", default_policy)


async def get_resource_policies(
    principal_arn: str, resource_actions: Dict[str, Dict[str, Any]], account: str
) -> Tuple[List[Dict], bool]:
    resource_policies: List[Dict] = []
    cross_account_request: bool = False
    for resource_name, resource_info in resource_actions.items():
        resource_account: str = resource_info.get("account", "")
        if resource_account and resource_account != account:
            # This is a cross-account request. Might need a resource policy.
            cross_account_request = True
            resource_type: str = resource_info.get("type", "")
            resource_region: str = resource_info.get("region", "")
            old_policy = await get_resource_policy(
                resource_account, resource_type, resource_name, resource_region
            )
            arns = resource_info.get("arns", [])
            actions = resource_info.get("actions", [])
            new_policy = await generate_updated_resource_policy(
                old_policy, principal_arn, arns, actions
            )

            result = {
                "resource": resource_name,
                "account": resource_account,
                "type": resource_type,
                "region": resource_region,
                "policy_document": new_policy,
            }
            resource_policies.append(result)

    return resource_policies, cross_account_request


async def generate_updated_resource_policy(
    existing: Dict,
    principal_arn: str,
    resource_arns: List[str],
    actions: List[str],
    include_resources: bool = True,
) -> Dict:
    """

    :param existing: Dict: the current existing policy document
    :param principal_arn: the Principal ARN which wants access to the resource
    :param resource_arns: the Resource ARNs
    :param actions: The list of Actions to be added
    :param include_resources: whether to include resources in the new statement or not
    :return: Dict: generated updated resource policy that includes a new statement for the listed actions
    """
    policy_dict = deepcopy(existing)
    new_statement = {
        "Effect": "Allow",
        "Principal": {"AWS": [principal_arn]},
        "Action": list(set(actions)),
    }
    if include_resources:
        new_statement["Resource"] = resource_arns
    policy_dict["Statement"].append(new_statement)
    return policy_dict


async def fetch_resource_details(
    account_id: str,
    resource_type: str,
    resource_name: str,
    region: str,
    path: str = None,
) -> dict:
    if resource_type == "s3":
        return await fetch_s3_bucket(account_id, resource_name)
    elif resource_type == "sqs":
        return await fetch_sqs_queue(account_id, region, resource_name)
    elif resource_type == "sns":
        return await fetch_sns_topic(account_id, region, resource_name)
    elif resource_type == "managed_policy":
        return await fetch_managed_policy_details(account_id, resource_name, path)
    else:
        return {}


async def fetch_managed_policy_details(
    account_id: str, resource_name: str, path: str = None
) -> Optional[Dict]:
    from consoleme.lib.policies import get_aws_config_history_url_for_resource

    if path:
        resource_name = path + "/" + resource_name
    policy_arn: str = f"arn:aws:iam::{account_id}:policy/{resource_name}"
    result: Dict = {}
    result["Policy"] = await sync_to_async(get_managed_policy_document)(
        policy_arn=policy_arn,
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        region=config.region,
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )
    policy_details = await sync_to_async(get_policy)(
        policy_arn=policy_arn,
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        region=config.region,
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )

    try:
        result["TagSet"] = policy_details["Policy"]["Tags"]
    except KeyError:
        result["TagSet"] = []
    result["config_timeline_url"] = await get_aws_config_history_url_for_resource(
        account_id,
        policy_arn,
        resource_name,
        "AWS::IAM::ManagedPolicy",
        region=config.region,
    )

    return result


async def fetch_assume_role_policy(role_arn: str) -> Optional[Dict]:
    account_id = role_arn.split(":")[4]
    role_name = role_arn.split("/")[-1]
    try:
        role = await fetch_role_details(account_id, role_name)
    except ClientError:
        # Role is most likely on an account that we do not have access to
        sentry_sdk.capture_exception()
        return None
    return role.assume_role_policy_document


async def fetch_sns_topic(account_id: str, region: str, resource_name: str) -> dict:
    from consoleme.lib.policies import get_aws_config_history_url_for_resource

    regions = await get_enabled_regions_for_account(account_id)
    if region not in regions:
        raise InvalidInvocationArgument(
            f"Region '{region}' is not valid region on account '{account_id}'."
        )

    arn: str = f"arn:aws:sns:{region}:{account_id}:{resource_name}"
    client = await sync_to_async(boto3_cached_conn)(
        "sns",
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        region=region,
        sts_client_kwargs=dict(
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        ),
        client_kwargs=config.get("boto3.client_kwargs", {}),
        retry_max_attempts=2,
    )

    result: Dict = await sync_to_async(get_topic_attributes)(
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        TopicArn=arn,
        region=region,
        sts_client_kwargs=dict(
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        ),
        client_kwargs=config.get("boto3.client_kwargs", {}),
        retry_max_attempts=2,
    )

    tags: Dict = await sync_to_async(client.list_tags_for_resource)(ResourceArn=arn)
    result["TagSet"] = tags["Tags"]
    if not isinstance(result["Policy"], dict):
        result["Policy"] = json.loads(result["Policy"])

    result["config_timeline_url"] = await get_aws_config_history_url_for_resource(
        account_id,
        arn,
        resource_name,
        "AWS::SNS::Topic",
        region=region,
    )
    return result


async def fetch_sqs_queue(account_id: str, region: str, resource_name: str) -> dict:
    from consoleme.lib.policies import get_aws_config_history_url_for_resource

    regions = await get_enabled_regions_for_account(account_id)
    if region not in regions:
        raise InvalidInvocationArgument(
            f"Region '{region}' is not valid region on account '{account_id}'."
        )

    queue_url: str = await sync_to_async(get_queue_url)(
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        region=region,
        QueueName=resource_name,
        sts_client_kwargs=dict(
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        ),
        client_kwargs=config.get("boto3.client_kwargs", {}),
        retry_max_attempts=2,
    )

    result: Dict = await sync_to_async(get_queue_attributes)(
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        region=region,
        QueueUrl=queue_url,
        AttributeNames=["All"],
        sts_client_kwargs=dict(
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        ),
        client_kwargs=config.get("boto3.client_kwargs", {}),
        retry_max_attempts=2,
    )

    tags: Dict = await sync_to_async(list_queue_tags)(
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        region=region,
        QueueUrl=queue_url,
        sts_client_kwargs=dict(
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        ),
        client_kwargs=config.get("boto3.client_kwargs", {}),
        retry_max_attempts=2,
    )
    result["TagSet"]: list = []
    result["QueueUrl"]: str = queue_url
    if tags:
        result["TagSet"] = [{"Key": k, "Value": v} for k, v in tags.items()]
    if result.get("CreatedTimestamp"):
        result["created_time"] = datetime.utcfromtimestamp(
            int(float(result["CreatedTimestamp"]))
        ).isoformat()
    if result.get("LastModifiedTimestamp"):
        result["updated_time"] = datetime.utcfromtimestamp(
            int(float(result["LastModifiedTimestamp"]))
        ).isoformat()
    # Unfortunately, the queue_url we get from our `get_queue_url` call above doesn't match the ID of the queue in
    # AWS Config, so we must hack our own.
    queue_url_manual = (
        f"https://sqs.{region}.amazonaws.com/{account_id}/{resource_name}"
    )
    result["config_timeline_url"] = await get_aws_config_history_url_for_resource(
        account_id,
        queue_url_manual,
        resource_name,
        "AWS::SQS::Queue",
        region=region,
    )
    return result


async def get_bucket_location_with_fallback(
    bucket_name: str, account_id: str, fallback_region: str = config.region
) -> str:
    try:
        bucket_location_res = await sync_to_async(get_bucket_location)(
            Bucket=bucket_name,
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            region=config.region,
            sts_client_kwargs=dict(
                region_name=config.region,
                endpoint_url=f"https://sts.{config.region}.amazonaws.com",
            ),
            client_kwargs=config.get("boto3.client_kwargs", {}),
            retry_max_attempts=2,
        )
        bucket_location = bucket_location_res.get("LocationConstraint", fallback_region)
        if not bucket_location:
            # API get_bucket_location returns None for buckets in us-east-1
            bucket_location = "us-east-1"
        if bucket_location == "EU":
            bucket_location = "eu-west-1"
        if bucket_location == "US":
            bucket_location = "us-east-1"
    except ClientError:
        bucket_location = fallback_region
        sentry_sdk.capture_exception()
    return bucket_location


async def fetch_s3_bucket(account_id: str, bucket_name: str) -> dict:
    """Fetch S3 Bucket and applicable policies

    :param account_id:
    :param bucket_name:
    :return:
    """

    from consoleme.lib.policies import get_aws_config_history_url_for_resource

    log_data: Dict = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "bucket_name": bucket_name,
        "account_id": account_id,
    }
    log.debug(log_data)
    created_time = None
    bucket_location = "us-east-1"

    try:
        bucket_resource = await sync_to_async(get_bucket_resource)(
            bucket_name,
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            region=config.region,
            sts_client_kwargs=dict(
                region_name=config.region,
                endpoint_url=f"https://sts.{config.region}.amazonaws.com",
            ),
            client_kwargs=config.get("boto3.client_kwargs", {}),
            retry_max_attempts=2,
        )
        created_time_stamp = bucket_resource.creation_date
        if created_time_stamp:
            created_time = created_time_stamp.isoformat()
    except ClientError:
        sentry_sdk.capture_exception()
    try:
        bucket_location = await get_bucket_location_with_fallback(
            bucket_name, account_id
        )
        policy: Dict = await sync_to_async(get_bucket_policy)(
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            region=bucket_location,
            Bucket=bucket_name,
            sts_client_kwargs=dict(
                region_name=config.region,
                endpoint_url=f"https://sts.{config.region}.amazonaws.com",
            ),
            client_kwargs=config.get("boto3.client_kwargs", {}),
            retry_max_attempts=2,
        )
    except ClientError as e:
        if "NoSuchBucketPolicy" in str(e):
            policy = {"Policy": "{}"}
        else:
            raise
    try:
        tags: Dict = await sync_to_async(get_bucket_tagging)(
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            region=bucket_location,
            Bucket=bucket_name,
            sts_client_kwargs=dict(
                region_name=config.region,
                endpoint_url=f"https://sts.{config.region}.amazonaws.com",
            ),
            client_kwargs=config.get("boto3.client_kwargs", {}),
            retry_max_attempts=2,
        )
    except ClientError as e:
        if "NoSuchTagSet" in str(e):
            tags = {"TagSet": []}
        else:
            raise

    result: Dict = {**policy, **tags, "created_time": created_time}
    result["config_timeline_url"] = await get_aws_config_history_url_for_resource(
        account_id,
        bucket_name,
        bucket_name,
        "AWS::S3::Bucket",
        region=bucket_location,
    )
    result["Policy"] = json.loads(result["Policy"])

    return result


async def raise_if_background_check_required_and_no_background_check(role, user):
    for compliance_account_id in config.get("aws.compliance_account_ids", []):
        if compliance_account_id == role.split(":")[4]:
            user_info = await auth.get_user_info(user, object=True)
            if not user_info.passed_background_check:
                function = f"{__name__}.{sys._getframe().f_code.co_name}"
                log_data: dict = {
                    "function": function,
                    "user": user,
                    "role": role,
                    "message": "User trying to access SEG role without background check",
                }
                log.error(log_data)
                stats.count(
                    f"{function}.access_denied_background_check_not_passed",
                    tags={"function": function, "user": user, "role": role},
                )
                raise BackgroundCheckNotPassedException(
                    config.get(
                        "aws.background_check_not_passed",
                        "You must have passed a background check to access role "
                        "{role}.",
                    ).format(role=role)
                )


def apply_managed_policy_to_role(
    role: Dict, policy_name: str, session_name: str
) -> bool:
    """
    Apply a managed policy to a role.
    :param role: An AWS role dictionary (from a boto3 get_role or get_account_authorization_details call)
    :param policy_name: Name of managed policy to add to role
    :param session_name: Name of session to assume role with. This is an identifier that will be logged in CloudTrail
    :return:
    """
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    log_data = {
        "function": function,
        "role": role,
        "policy_name": policy_name,
        "session_name": session_name,
    }
    account_id = role.get("Arn").split(":")[4]
    policy_arn = f"arn:aws:iam::{account_id}:policy/{policy_name}"
    client = boto3_cached_conn(
        "iam",
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        session_name=session_name,
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )

    client.attach_role_policy(RoleName=role.get("RoleName"), PolicyArn=policy_arn)
    log_data["message"] = "Applied managed policy to role"
    log.debug(log_data)
    stats.count(
        f"{function}.attach_role_policy",
        tags={"role": role.get("Arn"), "policy": policy_arn},
    )
    return True


async def delete_iam_user(account_id, iam_user_name, username) -> bool:
    """
    This function assumes the user has already been pre-authorized to delete an IAM user. it will detach all managed
    policies, delete all inline policies, delete all access keys, and finally delete the IAM user.

    :param account_id: Account ID that the IAM user is on
    :param iam_user_name: name of IAM user to delete
    :param username: actor's username
    :return:
    """
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Attempting to delete role",
        "account_id": account_id,
        "iam_user_name": iam_user_name,
        "user": username,
    }
    log.info(log_data)
    iam_user = await fetch_iam_user_details(account_id, iam_user_name)

    # Detach managed policies
    for policy in await sync_to_async(iam_user.attached_policies.all)():
        await sync_to_async(policy.load)()
        log.info(
            {
                **log_data,
                "message": "Detaching managed policy from user",
                "policy_arn": policy.arn,
            }
        )
        await sync_to_async(policy.detach_user)(UserName=iam_user)

    # Delete Inline policies
    for policy in await sync_to_async(iam_user.policies.all)():
        await sync_to_async(policy.load)()
        log.info(
            {
                **log_data,
                "message": "Deleting inline policy on user",
                "policy_name": policy.name,
            }
        )
        await sync_to_async(policy.delete)()

    log.info({**log_data, "message": "Performing access key deletion"})
    access_keys = iam_user.access_keys.all()
    for access_key in access_keys:
        access_key.delete()

    log.info({**log_data, "message": "Performing user deletion"})
    await sync_to_async(iam_user.delete)()
    stats.count(
        f"{log_data['function']}.success", tags={"iam_user_name": iam_user_name}
    )
    return True


async def delete_iam_role(account_id, role_name, username) -> bool:
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Attempting to delete role",
        "account_id": account_id,
        "role_name": role_name,
        "user": username,
    }
    log.info(log_data)
    role = await fetch_role_details(account_id, role_name)

    for instance_profile in await sync_to_async(role.instance_profiles.all)():
        await sync_to_async(instance_profile.load)()
        log.info(
            {
                **log_data,
                "message": "Removing and deleting instance profile from role",
                "instance_profile": instance_profile.name,
            }
        )
        await sync_to_async(instance_profile.remove_role)(RoleName=role.name)
        await sync_to_async(instance_profile.delete)()

    # Detach managed policies
    for policy in await sync_to_async(role.attached_policies.all)():
        await sync_to_async(policy.load)()
        log.info(
            {
                **log_data,
                "message": "Detaching managed policy from role",
                "policy_arn": policy.arn,
            }
        )
        await sync_to_async(policy.detach_role)(RoleName=role_name)

    # Delete Inline policies
    for policy in await sync_to_async(role.policies.all)():
        await sync_to_async(policy.load)()
        log.info(
            {
                **log_data,
                "message": "Deleting inline policy on role",
                "policy_name": policy.name,
            }
        )
        await sync_to_async(policy.delete)()

    log.info({**log_data, "message": "Performing role deletion"})
    await sync_to_async(role.delete)()
    stats.count(f"{log_data['function']}.success", tags={"role_name": role_name})


async def fetch_role_details(account_id, role_name):
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Attempting to fetch role details",
        "account": account_id,
        "role": role_name,
    }
    log.info(log_data)
    iam_resource = await sync_to_async(boto3_cached_conn)(
        "iam",
        service_type="resource",
        account_number=account_id,
        region=config.region,
        assume_role=config.get("policies.role_name"),
        session_name="fetch_role_details",
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )
    try:
        iam_role = await sync_to_async(iam_resource.Role)(role_name)
    except ClientError as ce:
        if ce.response["Error"]["Code"] == "NoSuchEntity":
            log_data["message"] = "Requested role doesn't exist"
            log.error(log_data)
        raise
    await sync_to_async(iam_role.load)()
    return iam_role


async def fetch_iam_user_details(account_id, iam_user_name):
    """
    Fetches details about an IAM user from AWS. If `policies.role_name` configuration
    is set, the hub (central) account ConsoleMeInstanceProfile role will assume the
    configured role to perform the action.

    :param account_id: account ID
    :param iam_user_name: IAM user name
    :return: iam_user resource
    """
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Attempting to fetch role details",
        "account": account_id,
        "iam_user_name": iam_user_name,
    }
    log.info(log_data)
    iam_resource = await sync_to_async(boto3_cached_conn)(
        "iam",
        service_type="resource",
        account_number=account_id,
        region=config.region,
        assume_role=config.get("policies.role_name"),
        session_name="fetch_iam_user_details",
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )
    try:
        iam_user = await sync_to_async(iam_resource.User)(iam_user_name)
    except ClientError as ce:
        if ce.response["Error"]["Code"] == "NoSuchEntity":
            log_data["message"] = "Requested user doesn't exist"
            log.error(log_data)
        raise
    await sync_to_async(iam_user.load)()
    return iam_user


async def create_iam_role(create_model: RoleCreationRequestModel, username):
    """
    Creates IAM role.
    :param create_model: RoleCreationRequestModel, which has the following attributes:
        account_id: destination account's ID
        role_name: destination role name
        description: optional string - description of the role
                     default: Role created by {username} through ConsoleMe
        instance_profile: optional boolean - whether to create an instance profile and attach it to the role or not
                     default: True
    :param username: username of user requesting action
    :return: results: - indicating the results of each action
    """
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Attempting to create role",
        "account_id": create_model.account_id,
        "role_name": create_model.role_name,
        "user": username,
    }
    log.info(log_data)

    default_trust_policy = config.get(
        "user_role_creator.default_trust_policy",
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "ec2.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        },
    )
    if default_trust_policy is None:
        raise MissingConfigurationValue(
            "Missing Default Assume Role Policy Configuration"
        )
    if create_model.description:
        description = create_model.description
    else:
        description = f"Role created by {username} through ConsoleMe"

    iam_client = await sync_to_async(boto3_cached_conn)(
        "iam",
        service_type="client",
        account_number=create_model.account_id,
        region=config.region,
        assume_role=config.get("policies.role_name"),
        session_name=sanitize_session_name("create_role_" + username),
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )
    results = {"errors": 0, "role_created": "false", "action_results": []}
    try:
        await sync_to_async(iam_client.create_role)(
            RoleName=create_model.role_name,
            AssumeRolePolicyDocument=json.dumps(default_trust_policy),
            Description=description,
            Tags=[],
        )
        results["action_results"].append(
            {
                "status": "success",
                "message": f"Role arn:aws:iam::{create_model.account_id}:role/{create_model.role_name} "
                f"successfully created",
            }
        )
        results["role_created"] = "true"
    except Exception as e:
        log_data["message"] = "Exception occurred creating role"
        log_data["error"] = str(e)
        log.error(log_data, exc_info=True)
        results["action_results"].append(
            {
                "status": "error",
                "message": f"Error creating role {create_model.role_name} in account {create_model.account_id}:"
                + str(e),
            }
        )
        results["errors"] += 1
        sentry_sdk.capture_exception()
        # Since we were unable to create the role, no point continuing, just return
        return results

    # If here, role has been successfully created, add status updates for each action
    results["action_results"].append(
        {
            "status": "success",
            "message": "Successfully added default Assume Role Policy Document",
        }
    )
    results["action_results"].append(
        {
            "status": "success",
            "message": "Successfully added description: " + description,
        }
    )

    # Create instance profile and attach if specified
    if create_model.instance_profile:
        try:
            await sync_to_async(iam_client.create_instance_profile)(
                InstanceProfileName=create_model.role_name
            )
            await sync_to_async(iam_client.add_role_to_instance_profile)(
                InstanceProfileName=create_model.role_name,
                RoleName=create_model.role_name,
            )
            results["action_results"].append(
                {
                    "status": "success",
                    "message": f"Successfully added instance profile {create_model.role_name} to role "
                    f"{create_model.role_name}",
                }
            )
        except Exception as e:
            log_data[
                "message"
            ] = "Exception occurred creating/attaching instance profile"
            log_data["error"] = str(e)
            log.error(log_data, exc_info=True)
            sentry_sdk.capture_exception()
            results["action_results"].append(
                {
                    "status": "error",
                    "message": f"Error creating/attaching instance profile {create_model.role_name} to role: "
                    + str(e),
                }
            )
            results["errors"] += 1

    stats.count(
        f"{log_data['function']}.success", tags={"role_name": create_model.role_name}
    )
    log_data["message"] = "Successfully created role"
    log.info(log_data)
    # Force caching of role
    try:
        aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
        role_arn = (
            f"arn:aws:iam::{create_model.account_id}:role/{create_model.role_name}"
        )
        await aws.fetch_iam_role(create_model.account_id, role_arn, force_refresh=True)
    except Exception as e:
        log.error({**log_data, "message": "Unable to cache role", "error": str(e)})
        sentry_sdk.capture_exception()
    return results


async def clone_iam_role(clone_model: CloneRoleRequestModel, username):
    """
    Clones IAM role within same account or across account, always creating and attaching instance profile if one exists
    on the source role.
    ;param username: username of user requesting action
    ;:param clone_model: CloneRoleRequestModel, which has the following attributes:
        account_id: source role's account ID
        role_name: source role's name
        dest_account_id: destination role's account ID (may be same as account_id)
        dest_role_name: destination role's name
        clone_options: dict to indicate what to copy when cloning:
            assume_role_policy: bool
                default: False - uses default ConsoleMe AssumeRolePolicy
            tags: bool
                default: False - defaults to no tags
            copy_description: bool
                default: False - defaults to copying provided description or default description
            description: string
                default: "Role cloned via ConsoleMe by `username` from `arn:aws:iam::<account_id>:role/<role_name>`
                if copy_description is True, then description is ignored
            inline_policies: bool
                default: False - defaults to no inline policies
            managed_policies: bool
                default: False - defaults to no managed policies
    :return: results: - indicating the results of each action
    """

    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Attempting to clone role",
        "account_id": clone_model.account_id,
        "role_name": clone_model.role_name,
        "dest_account_id": clone_model.dest_account_id,
        "dest_role_name": clone_model.dest_role_name,
        "user": username,
    }
    log.info(log_data)
    role = await fetch_role_details(clone_model.account_id, clone_model.role_name)

    default_trust_policy = config.get("user_role_creator.default_trust_policy")
    trust_policy = (
        role.assume_role_policy_document
        if clone_model.options.assume_role_policy
        else default_trust_policy
    )
    if trust_policy is None:
        raise MissingConfigurationValue(
            "Missing Default Assume Role Policy Configuration"
        )

    if (
        clone_model.options.copy_description
        and role.description is not None
        and role.description != ""
    ):
        description = role.description
    elif (
        clone_model.options.description is not None
        and clone_model.options.description != ""
    ):
        description = clone_model.options.description
    else:
        description = f"Role cloned via ConsoleMe by {username} from {role.arn}"

    tags = role.tags if clone_model.options.tags and role.tags else []

    iam_client = await sync_to_async(boto3_cached_conn)(
        "iam",
        service_type="client",
        account_number=clone_model.dest_account_id,
        region=config.region,
        assume_role=config.get("policies.role_name"),
        session_name=sanitize_session_name("clone_role_" + username),
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )
    results = {"errors": 0, "role_created": "false", "action_results": []}
    try:
        await sync_to_async(iam_client.create_role)(
            RoleName=clone_model.dest_role_name,
            AssumeRolePolicyDocument=json.dumps(trust_policy),
            Description=description,
            Tags=tags,
        )
        results["action_results"].append(
            {
                "status": "success",
                "message": f"Role arn:aws:iam::{clone_model.dest_account_id}:role/{clone_model.dest_role_name} "
                f"successfully created",
            }
        )
        results["role_created"] = "true"
    except Exception as e:
        log_data["message"] = "Exception occurred creating cloned role"
        log_data["error"] = str(e)
        log.error(log_data, exc_info=True)
        results["action_results"].append(
            {
                "status": "error",
                "message": f"Error creating role {clone_model.dest_role_name} in account {clone_model.dest_account_id}:"
                + str(e),
            }
        )
        results["errors"] += 1
        sentry_sdk.capture_exception()
        # Since we were unable to create the role, no point continuing, just return
        return results

    if clone_model.options.tags:
        results["action_results"].append(
            {"status": "success", "message": "Successfully copied tags"}
        )
    if clone_model.options.assume_role_policy:
        results["action_results"].append(
            {
                "status": "success",
                "message": "Successfully copied Assume Role Policy Document",
            }
        )
    else:
        results["action_results"].append(
            {
                "status": "success",
                "message": "Successfully added default Assume Role Policy Document",
            }
        )
    if (
        clone_model.options.copy_description
        and role.description is not None
        and role.description != ""
    ):
        results["action_results"].append(
            {"status": "success", "message": "Successfully copied description"}
        )
    elif clone_model.options.copy_description:
        results["action_results"].append(
            {
                "status": "error",
                "message": "Failed to copy description, so added default description: "
                + description,
            }
        )
    else:
        results["action_results"].append(
            {
                "status": "success",
                "message": "Successfully added description: " + description,
            }
        )
    # Create instance profile and attach if it exists in source role
    if len(list(await sync_to_async(role.instance_profiles.all)())) > 0:
        try:
            await sync_to_async(iam_client.create_instance_profile)(
                InstanceProfileName=clone_model.dest_role_name
            )
            await sync_to_async(iam_client.add_role_to_instance_profile)(
                InstanceProfileName=clone_model.dest_role_name,
                RoleName=clone_model.dest_role_name,
            )
            results["action_results"].append(
                {
                    "status": "success",
                    "message": f"Successfully added instance profile {clone_model.dest_role_name} to role "
                    f"{clone_model.dest_role_name}",
                }
            )
        except Exception as e:
            log_data[
                "message"
            ] = "Exception occurred creating/attaching instance profile"
            log_data["error"] = str(e)
            log.error(log_data, exc_info=True)
            sentry_sdk.capture_exception()
            results["action_results"].append(
                {
                    "status": "error",
                    "message": f"Error creating/attaching instance profile {clone_model.dest_role_name} to role: "
                    + str(e),
                }
            )
            results["errors"] += 1

    # other optional attributes to copy over after role has been successfully created

    cloned_role = await fetch_role_details(
        clone_model.dest_account_id, clone_model.dest_role_name
    )

    # Copy inline policies
    if clone_model.options.inline_policies:
        for src_policy in await sync_to_async(role.policies.all)():
            await sync_to_async(src_policy.load)()
            try:
                dest_policy = await sync_to_async(cloned_role.Policy)(src_policy.name)
                await sync_to_async(dest_policy.put)(
                    PolicyDocument=json.dumps(src_policy.policy_document)
                )
                results["action_results"].append(
                    {
                        "status": "success",
                        "message": f"Successfully copied inline policy {src_policy.name}",
                    }
                )
            except Exception as e:
                log_data["message"] = "Exception occurred copying inline policy"
                log_data["error"] = str(e)
                log.error(log_data, exc_info=True)
                sentry_sdk.capture_exception()
                results["action_results"].append(
                    {
                        "status": "error",
                        "message": f"Error copying inline policy {src_policy.name}: "
                        + str(e),
                    }
                )
                results["errors"] += 1

    # Copy managed policies
    if clone_model.options.managed_policies:
        for src_policy in await sync_to_async(role.attached_policies.all)():
            await sync_to_async(src_policy.load)()
            dest_policy_arn = src_policy.arn.replace(
                clone_model.account_id, clone_model.dest_account_id
            )
            try:
                await sync_to_async(cloned_role.attach_policy)(
                    PolicyArn=dest_policy_arn
                )
                results["action_results"].append(
                    {
                        "status": "success",
                        "message": f"Successfully attached managed policy {src_policy.arn} as {dest_policy_arn}",
                    }
                )
            except Exception as e:
                log_data["message"] = "Exception occurred copying managed policy"
                log_data["error"] = str(e)
                log.error(log_data, exc_info=True)
                sentry_sdk.capture_exception()
                results["action_results"].append(
                    {
                        "status": "error",
                        "message": f"Error attaching managed policy {dest_policy_arn}: "
                        + str(e),
                    }
                )
                results["errors"] += 1

    stats.count(
        f"{log_data['function']}.success", tags={"role_name": clone_model.role_name}
    )
    log_data["message"] = "Successfully cloned role"
    log.info(log_data)
    return results


def role_has_tag(role: Dict, key: str, value: Optional[str] = None) -> bool:
    """
    Checks a role dictionary and determine of the role has the specified tag. If `value` is passed,
    This function will only return true if the tag's value matches the `value` variable.
    :param role: An AWS role dictionary (from a boto3 get_role or get_account_authorization_details call)
    :param key: key of the tag
    :param value: optional value of the tag
    :return:
    """
    for tag in role.get("Tags", []):
        if tag.get("Key") == key:
            if not value or tag.get("Value") == value:
                return True
    return False


def role_has_managed_policy(role: Dict, managed_policy_name: str) -> bool:
    """
    Checks a role dictionary to determine if a managed policy is attached
    :param role: An AWS role dictionary (from a boto3 get_role or get_account_authorization_details call)
    :param managed_policy_name: the name of the managed policy
    :return:
    """

    for managed_policy in role.get("AttachedManagedPolicies", []):
        if managed_policy.get("PolicyName") == managed_policy_name:
            return True
    return False


def role_newer_than_x_days(role: Dict, days: int) -> bool:
    """
    Checks a role dictionary to determine if it is newer than the specified number of days
    :param role:  An AWS role dictionary (from a boto3 get_role or get_account_authorization_details call)
    :param days: number of days
    :return:
    """
    if isinstance(role.get("CreateDate"), str):
        role["CreateDate"] = parse(role.get("CreateDate"))
    role_age = datetime.now(tz=pytz.utc) - role.get("CreateDate")
    if role_age.days < days:
        return True
    return False


def is_role_instance_profile(role: Dict) -> bool:
    """
    Checks a role naively to determine if it is associate with an instance profile.
    We only check by name, and not the actual attached instance profiles.
    :param role: An AWS role dictionary (from a boto3 get_role or get_account_authorization_details call)
    :return:
    """
    return role.get("RoleName").endswith("InstanceProfile")


def get_region_from_arn(arn):
    """Given an ARN, return the region in the ARN, if it is available. In certain cases like S3 it is not"""
    result = parse_arn(arn)
    # Support S3 buckets with no values under region
    if result["region"] is None:
        result = ""
    else:
        result = result["region"]
    return result


def get_resource_from_arn(arn):
    """Given an ARN, parse it according to ARN namespacing and return the resource. See
    http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html for more details on ARN namespacing.
    """
    result = parse_arn(arn)
    return result["resource"]


def get_service_from_arn(arn):
    """Given an ARN string, return the service"""
    result = parse_arn(arn)
    return result["service"]


async def get_enabled_regions_for_account(account_id: str) -> Set[str]:
    """
    Returns a list of regions enabled for an account based on an EC2 Describe Regions call. Can be overridden with a
    global configuration of static regions (Configuration key: `celery.sync_regions`), or a configuration of specific
    regions per account (Configuration key:  `get_enabled_regions_for_account.{account_id}`)
    """
    enabled_regions_for_account = config.get(
        f"get_enabled_regions_for_account.{account_id}"
    )
    if enabled_regions_for_account:
        return enabled_regions_for_account

    celery_sync_regions = config.get("celery.sync_regions", [])
    if celery_sync_regions:
        return celery_sync_regions

    client = await sync_to_async(boto3_cached_conn)(
        "ec2",
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        read_only=True,
        retry_max_attempts=2,
        client_kwargs=config.get("boto3.client_kwargs", {}),
    )

    regions = await sync_to_async(client.describe_regions)()
    return {r["RegionName"] for r in regions["Regions"]}


async def access_analyzer_validate_policy(
    policy: str, log_data, policy_type: str = "IDENTITY_POLICY"
) -> List[Dict[str, Any]]:
    try:
        enhanced_findings = []
        client = await sync_to_async(boto3.client)(
            "accessanalyzer",
            region_name=config.region,
            **config.get("boto3.client_kwargs", {}),
        )
        access_analyzer_response = await sync_to_async(client.validate_policy)(
            policyDocument=policy,
            policyType=policy_type,  # ConsoleMe only supports identity policy analysis currently
        )
        for finding in access_analyzer_response.get("findings", []):
            for location in finding.get("locations", []):
                enhanced_findings.append(
                    {
                        "issue": finding.get("issueCode"),
                        "detail": "",
                        "location": {
                            "line": location.get("span", {})
                            .get("start", {})
                            .get("line"),
                            "column": location.get("span", {})
                            .get("start", {})
                            .get("column"),
                            "filepath": None,
                        },
                        "severity": finding.get("findingType"),
                        "title": finding.get("issueCode"),
                        "description": finding.get("findingDetails"),
                    }
                )
        return enhanced_findings
    except (ParamValidationError, ClientError) as e:
        log.error(
            {
                **log_data,
                "function": f"{__name__}.{sys._getframe().f_code.co_name}",
                "message": "Error retrieving Access Analyzer data",
                "policy": policy,
                "error": str(e),
            }
        )
        sentry_sdk.capture_exception()
        return []


async def parliament_validate_iam_policy(policy: str) -> List[Dict[str, Any]]:
    analyzed_policy = await sync_to_async(analyze_policy_string)(policy)
    findings = analyzed_policy.findings

    enhanced_findings = []

    for finding in findings:
        enhanced_finding = await sync_to_async(enhance_finding)(finding)
        enhanced_findings.append(
            {
                "issue": enhanced_finding.issue,
                "detail": json.dumps(enhanced_finding.detail),
                "location": enhanced_finding.location,
                "severity": enhanced_finding.severity,
                "title": enhanced_finding.title,
                "description": enhanced_finding.description,
            }
        )
    return enhanced_findings


async def validate_iam_policy(policy: str, log_data: Dict):
    parliament_findings: List = await parliament_validate_iam_policy(policy)
    access_analyzer_findings: List = await access_analyzer_validate_policy(
        policy, log_data, policy_type="IDENTITY_POLICY"
    )
    return parliament_findings + access_analyzer_findings


async def get_all_scps(force_sync=False) -> Dict[str, List[ServiceControlPolicyModel]]:
    """Retrieve a dictionary containing all Service Control Policies across organizations

    Args:
        force_sync: force a cache update
    """
    redis_key = config.get(
        "cache_scps_across_organizations.redis.key.all_scps_key", "ALL_AWS_SCPS"
    )
    scps = await retrieve_json_data_from_redis_or_s3(
        redis_key,
        s3_bucket=config.get("cache_scps_across_organizations.s3.bucket"),
        s3_key=config.get(
            "cache_scps_across_organizations.s3.file", "scps/cache_scps_v1.json.gz"
        ),
        default={},
        max_age=86400,
    )
    if force_sync or not scps:
        scps = await cache_all_scps()
    scp_models = {}
    for account, org_scps in scps.items():
        scp_models[account] = [ServiceControlPolicyModel(**scp) for scp in org_scps]
    return scp_models


async def cache_all_scps() -> Dict[str, Any]:
    """Store a dictionary of all Service Control Policies across organizations in the cache"""
    all_scps = {}
    for organization in config.get("cache_accounts_from_aws_organizations", []):
        org_account_id = organization.get("organizations_master_account_id")
        role_to_assume = organization.get(
            "organizations_master_role_to_assume", config.get("policies.role_name")
        )
        if not org_account_id:
            raise MissingConfigurationValue(
                "Your AWS Organizations Master Account ID is not specified in configuration. "
                "Unable to sync accounts from "
                "AWS Organizations"
            )

        if not role_to_assume:
            raise MissingConfigurationValue(
                "ConsoleMe doesn't know what role to assume to retrieve account information "
                "from AWS Organizations. please set the appropriate configuration value."
            )
        org_scps = await retrieve_scps_for_organization(
            org_account_id, role_to_assume=role_to_assume, region=config.region
        )
        all_scps[org_account_id] = org_scps
    redis_key = config.get(
        "cache_scps_across_organizations.redis.key.all_scps_key", "ALL_AWS_SCPS"
    )
    s3_bucket = None
    s3_key = None
    if config.region == config.get("celery.active_region", config.region) or config.get(
        "environment"
    ) in ["dev", "test"]:
        s3_bucket = config.get("cache_scps_across_organizations.s3.bucket")
        s3_key = config.get(
            "cache_scps_across_organizations.s3.file", "scps/cache_scps_v1.json.gz"
        )
    await store_json_results_in_redis_and_s3(
        all_scps,
        redis_key=redis_key,
        s3_bucket=s3_bucket,
        s3_key=s3_key,
    )
    return all_scps


async def get_org_structure(force_sync=False) -> Dict[str, Any]:
    """Retrieve a dictionary containing the organization structure

    Args:
        force_sync: force a cache update
    """
    redis_key = config.get(
        "cache_organization_structure.redis.key.org_structure_key", "AWS_ORG_STRUCTURE"
    )
    org_structure = await retrieve_json_data_from_redis_or_s3(
        redis_key,
        s3_bucket=config.get("cache_organization_structure.s3.bucket"),
        s3_key=config.get(
            "cache_organization_structure.s3.file",
            "scps/cache_org_structure_v1.json.gz",
        ),
        default={},
    )
    if force_sync or not org_structure:
        org_structure = await cache_org_structure()
    return org_structure


async def cache_org_structure() -> Dict[str, Any]:
    """Store a dictionary of the organization structure in the cache"""
    all_org_structure = {}
    for organization in config.get("cache_accounts_from_aws_organizations", []):
        org_account_id = organization.get("organizations_master_account_id")
        role_to_assume = organization.get(
            "organizations_master_role_to_assume", config.get("policies.role_name")
        )
        if not org_account_id:
            raise MissingConfigurationValue(
                "Your AWS Organizations Master Account ID is not specified in configuration. "
                "Unable to sync accounts from "
                "AWS Organizations"
            )

        if not role_to_assume:
            raise MissingConfigurationValue(
                "ConsoleMe doesn't know what role to assume to retrieve account information "
                "from AWS Organizations. please set the appropriate configuration value."
            )
        org_structure = await retrieve_org_structure(
            org_account_id, region=config.region
        )
        all_org_structure.update(org_structure)
    redis_key = config.get(
        "cache_organization_structure.redis.key.org_structure_key", "AWS_ORG_STRUCTURE"
    )
    s3_bucket = None
    s3_key = None
    if config.region == config.get("celery.active_region", config.region) or config.get(
        "environment"
    ) in ["dev", "test"]:
        s3_bucket = config.get("cache_organization_structure.s3.bucket")
        s3_key = config.get(
            "cache_organization_structure.s3.file",
            "scps/cache_org_structure_v1.json.gz",
        )
    await store_json_results_in_redis_and_s3(
        all_org_structure,
        redis_key=redis_key,
        s3_bucket=s3_bucket,
        s3_key=s3_key,
    )
    return all_org_structure


async def _is_member_of_ou(
    identifier: str, ou: Dict[str, Any]
) -> Tuple[bool, Set[str]]:
    """Recursively walk org structure to determine if the account or OU is in the org and, if so, return all OUs of which the account or OU is a member

    Args:
        identifier: AWS account or OU ID
        ou: dictionary representing the organization/organizational unit structure to search
    """
    found = False
    ou_path = set()
    for child in ou.get("Children", []):
        if child.get("Id") == identifier:
            found = True
        elif child.get("Type") == "ORGANIZATIONAL_UNIT":
            found, ou_path = await _is_member_of_ou(identifier, child)
        if found:
            ou_path.add(ou.get("Id"))
            break
    return found, ou_path


async def get_organizational_units_for_account(identifier: str) -> Set[str]:
    """Return a set of Organizational Unit IDs for a given account or OU ID

    Args:
        identifier: AWS account or OU ID
    """
    all_orgs = await get_org_structure()
    organizational_units = set()
    for org_id, org_structure in all_orgs.items():
        found, organizational_units = await _is_member_of_ou(identifier, org_structure)
        if found:
            break
    if not organizational_units:
        log.warning("could not find account in organization")
    return organizational_units


async def _scp_targets_account_or_ou(
    scp: ServiceControlPolicyModel, identifier: str, organizational_units: Set[str]
) -> bool:
    """Return True if the provided SCP targets the account or OU identifier provided

    Args:
        scp: Service Control Policy whose targets we check
        identifier: AWS account or OU ID
        organizational_units: set of IDs for OUs of which the identifier is a member
    """
    for target in scp.targets:
        if target.target_id == identifier or target.target_id in organizational_units:
            return True
    return False


async def get_scps_for_account_or_ou(identifier: str) -> ServiceControlPolicyArrayModel:
    """Retrieve a list of Service Control Policies for the account or OU specified by the identifier

    Args:
        identifier: AWS account or OU ID
    """
    all_scps = await get_all_scps()
    account_ous = await get_organizational_units_for_account(identifier)
    scps_for_account = []
    for org_account_id, scps in all_scps.items():
        # Iterate through each org's SCPs and see if the provided account_id is in the targets
        for scp in scps:
            if await _scp_targets_account_or_ou(scp, identifier, account_ous):
                scps_for_account.append(scp)
    scps = ServiceControlPolicyArrayModel(__root__=scps_for_account)
    return scps


async def minimize_iam_policy_statements(
    inline_iam_policy_statements: List[Dict], disregard_sid=True
) -> List[Dict]:
    """
    Minimizes a list of inline IAM policy statements.

    1. Policies that are identical except for the resources will have the resources merged into a single statement
    with the same actions, effects, conditions, etc.

    2. Policies that have an identical resource, but different actions, will be combined if the rest of the policy
    is identical.
    :param inline_iam_policy_statements: A list of IAM policy statement dictionaries
    :return: A potentially more compact list of IAM policy statement dictionaries
    """
    exclude_ids = []
    minimized_statements = []

    inline_iam_policy_statements = await normalize_policies(
        inline_iam_policy_statements
    )

    for i in range(len(inline_iam_policy_statements)):
        inline_iam_policy_statement = inline_iam_policy_statements[i]
        if disregard_sid:
            inline_iam_policy_statement.pop("Sid", None)
        if i in exclude_ids:
            # We've already combined this policy with another. Ignore it.
            continue
        for j in range(i + 1, len(inline_iam_policy_statements)):
            if j in exclude_ids:
                # We've already combined this policy with another. Ignore it.
                continue
            inline_iam_policy_statement_to_compare = inline_iam_policy_statements[j]
            if disregard_sid:
                inline_iam_policy_statement_to_compare.pop("Sid", None)
            # Check to see if policy statements are identical except for a given element. Merge the policies
            # if possible.
            for element in [
                "Resource",
                "Action",
                "NotAction",
                "NotResource",
                "NotPrincipal",
            ]:
                if not (
                    inline_iam_policy_statement.get(element)
                    or inline_iam_policy_statement_to_compare.get(element)
                ):
                    # This function won't handle `Condition`.
                    continue
                diff = DeepDiff(
                    inline_iam_policy_statement,
                    inline_iam_policy_statement_to_compare,
                    ignore_order=True,
                    exclude_paths=[f"root['{element}']"],
                )
                if not diff:
                    exclude_ids.append(j)
                    # Policy can be minimized
                    inline_iam_policy_statement[element] = sorted(
                        list(
                            set(
                                inline_iam_policy_statement[element]
                                + inline_iam_policy_statement_to_compare[element]
                            )
                        )
                    )
                    break

    for i in range(len(inline_iam_policy_statements)):
        if i not in exclude_ids:
            inline_iam_policy_statements[i] = sort_dict(inline_iam_policy_statements[i])
            minimized_statements.append(inline_iam_policy_statements[i])
    # TODO(cccastrapel): Intelligently combine actions and/or resources if they include wildcards
    minimized_statements = await normalize_policies(minimized_statements)
    return minimized_statements


async def normalize_policies(policies: List[Any]) -> List[Any]:
    """
    Normalizes policy statements to ensure appropriate AWS policy elements are lists (such as actions and resources),
    lowercase, and sorted. It will remove duplicate entries and entries that are superseded by other elements.
    """

    for policy in policies:
        for element in [
            "Resource",
            "Action",
            "NotAction",
            "NotResource",
            "NotPrincipal",
        ]:
            if not policy.get(element):
                continue
            if isinstance(policy.get(element), str):
                policy[element] = [policy[element]]
            # Policy elements can be lowercased, except for resources. Some resources
            # (such as IAM roles) are case sensitive
            if element in ["Resource", "NotResource", "NotPrincipal"]:
                policy[element] = list(set(policy[element]))
            else:
                policy[element] = list(set([x.lower() for x in policy[element]]))
            modified_elements = set()
            for i in range(len(policy[element])):
                matched = False
                # Sorry for the magic. this is iterating through all elements of a list that aren't the current element
                for compare_value in policy[element][:i] + policy[element][(i + 1) :]:
                    if fnmatch.fnmatch(policy[element][i], compare_value):
                        matched = True
                        break
                if not matched:
                    modified_elements.add(policy[element][i])
            policy[element] = sorted(modified_elements)
    return policies


def allowed_to_sync_role(
    role_arn: str, role_tags: List[Optional[Dict[str, str]]]
) -> bool:
    """
    This function determines whether ConsoleMe is allowed to sync or otherwise manipulate an IAM role. By default,
    ConsoleMe will sync all roles that it can get its grubby little hands on. However, ConsoleMe administrators can tell
    ConsoleMe to only sync roles with either 1) Specific ARNs, or 2) Specific tag key/value pairs. All configured tags
    must exist on the role for ConsoleMe to sync it.

    Here's an example configuration for a tag-based restriction:

    ```
    roles:
      allowed_tags:
        tag1: value1
        tag2: value2
    ```

    And another one for an ARN-based restriction:

    ```
    roles:
      allowed_arns:
        - arn:aws:iam::111111111111:role/role-name-here-1
        - arn:aws:iam::111111111111:role/role-name-here-2
        - arn:aws:iam::111111111111:role/role-name-here-3
        - arn:aws:iam::222222222222:role/role-name-here-1
        - arn:aws:iam::333333333333:role/role-name-here-1
    ```

    :param
        arn: The AWS role arn
        role_tags: A dictionary of role tags

    :return: boolean specifying whether ConsoleMe is allowed to sync / access the role
    """
    allowed_tags = config.get("roles.allowed_tags", {})
    allowed_arns = config.get("roles.allowed_arns", [])
    if not allowed_tags and not allowed_arns:
        return True

    if role_arn in allowed_arns:
        return True

    # Convert list of role tag dicts to a single key/value dict of tags
    # ex:
    # role_tags = [{'Key': 'consoleme-authorized', 'Value': 'consoleme_admins'},
    # {'Key': 'Description', 'Value': 'ConsoleMe OSS Demo Role'}]
    # so: actual_tags = {'consoleme-authorized': 'consoleme_admins', 'Description': 'ConsoleMe OSS Demo Role'}
    actual_tags = {
        d["Key"]: d["Value"] for d in role_tags
    }  # Convert List[Dicts] to 1 Dict

    # All configured allowed_tags must exist in the role's actual_tags for this condition to pass
    if allowed_tags and allowed_tags.items() <= actual_tags.items():
        return True
    return False


def get_aws_principal_owner(role_details: Dict[str, Any]) -> Optional[str]:
    """
    Identifies the owning user/group of an AWS principal based on one or more trusted and configurable principal tags.
    `owner` is used to notify application owners of permission problems with their detected AWS principals or resources
    if another identifier (ie: session name) for a principal doesn't point to a specific user for notification.

    :return: owner: str
    """
    owner = None
    owner_tag_names = config.get("aws.tags.owner", [])
    if not owner_tag_names:
        return owner
    if isinstance(owner_tag_names, str):
        owner_tag_names = [owner_tag_names]
    role_tags = role_details.get("Tags")
    for owner_tag_name in owner_tag_names:
        for role_tag in role_tags:
            if role_tag["Key"] == owner_tag_name:
                return role_tag["Value"]
    return owner


async def resource_arn_known_in_aws_config(
    resource_arn: str,
    run_query: bool = True,
    run_query_with_aggregator: bool = True,
    expiration_seconds: int = config.get(
        "aws.resource_arn_known_in_aws_config.expiration_seconds", 3600
    ),
) -> bool:
    """
    Determines if the resource ARN is known in AWS Config. AWS config does not store all resource
    types, nor will it account for cross-organizational resources, so the result of this function shouldn't be used
    to determine if a resource "exists" or not.

    A more robust approach is determining the resource type and querying AWS API directly to see if it exists, but this
    requires a lot of code.

    Note: This data may be stale by ~ 1 hour and 15 minutes (local results caching + typical AWS config delay)

    :param expiration_seconds: Number of seconds to consider stored result expired
    :param resource_arn: ARN of the resource we want to look up
    :param run_query: Should we run an AWS config query if we're not able to find the resource in our AWS Config cache?
    :param run_query_with_aggregator: Should we run the AWS Config query on our AWS Config aggregator?
    :return:
    """
    known_arn = False
    if not resource_arn.startswith("arn:aws:"):
        return known_arn

    resources_from_aws_config_redis_key: str = config.get(
        "aws_config_cache.redis_key", "AWSCONFIG_RESOURCE_CACHE"
    )

    if red.exists(resources_from_aws_config_redis_key) and red.hget(
        resources_from_aws_config_redis_key, resource_arn
    ):
        return True

    resource_arn_exists_temp_matches_redis_key: str = config.get(
        "resource_arn_known_in_aws_config.redis.temp_matches_key",
        "TEMP_QUERIED_RESOURCE_ARN_CACHE",
    )

    # To prevent repetitive queries against AWS config, first see if we've already ran a query recently
    result = await redis_hgetex(
        resource_arn_exists_temp_matches_redis_key, resource_arn
    )
    if result:
        return result["known"]

    if not run_query:
        return False

    r = await sync_to_async(query)(
        f"select arn where arn = '{resource_arn}'",
        use_aggregator=run_query_with_aggregator,
    )
    if r:
        known_arn = True
    # To prevent future repetitive queries on AWS Config, set our result in Redis with an expiration
    await redis_hsetex(
        resource_arn_exists_temp_matches_redis_key,
        resource_arn,
        {"known": known_arn},
        expiration_seconds=expiration_seconds,
    )

    return known_arn


async def simulate_iam_principal_action(
    principal_arn,
    action,
    resource_arn,
    source_ip,
    expiration_seconds: int = config.get(
        "aws.simulate_iam_principal_action.expiration_seconds", 3600
    ),
):
    """
    Simulates an IAM principal action affecting a resource

    :return:
    """
    # simulating IAM principal policies is expensive.
    # Temporarily cache and return results by principal_arn, action, and resource_arn. We don't consider source_ip
    # when caching because it could vary greatly for application roles running on multiple instances/containers.
    resource_arn_exists_temp_matches_redis_key: str = config.get(
        "resource_arn_known_in_aws_config.redis.temp_matches_key",
        "TEMP_POLICY_SIMULATION_CACHE",
    )

    cache_key = f"{principal_arn}-{action}-{resource_arn}"
    result = await redis_hgetex(resource_arn_exists_temp_matches_redis_key, cache_key)
    if result:
        return result

    ip_regex = r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
    context_entries = []
    if source_ip and re.match(ip_regex, source_ip):
        context_entries.append(
            {
                "ContextKeyName": "aws:SourceIp",
                "ContextKeyValues": [source_ip],
                "ContextKeyType": "ip",
            }
        )
    account_id = principal_arn.split(":")[4]
    client = await sync_to_async(boto3_cached_conn)(
        "iam",
        account_number=account_id,
        assume_role=config.get("policies.role_name"),
        sts_client_kwargs=dict(
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        ),
        retry_max_attempts=2,
    )
    try:
        response = await sync_to_async(client.simulate_principal_policy)(
            PolicySourceArn=principal_arn,
            ActionNames=[
                action,
            ],
            ResourceArns=[
                resource_arn,
            ],
            # TODO: Attach resource policy when discoverable
            # ResourcePolicy='string',
            # TODO: Attach Account ID of resource
            # ResourceOwner='string',
            ContextEntries=context_entries,
            MaxItems=100,
        )

        await redis_hsetex(
            resource_arn_exists_temp_matches_redis_key,
            resource_arn,
            response["EvaluationResults"],
            expiration_seconds=expiration_seconds,
        )
    except Exception:
        sentry_sdk.capture_exception()
        return None
    return response["EvaluationResults"]


async def get_iam_principal_owner(arn: str, aws: Any) -> Optional[str]:
    principal_details = {}
    principal_type = arn.split(":")[-1].split("/")[0]
    account_id = arn.split(":")[4]
    # trying to find principal for subsequent queries
    if principal_type == "role":
        principal_details = await aws().fetch_iam_role(account_id, arn)
    elif principal_type == "user":
        principal_details = await aws().fetch_iam_user(account_id, arn)
    return principal_details.get("owner")


def sanitize_session_name(unsanitized_session_name):
    """
    This function sanitizes the session name typically passed in an assume_role call, to verify that it's
    """
    valid_characters_re = re.compile(r"[\w+=,.@-]")

    sanitized_session_name = ""
    max_length = 64  # Session names have a length limit of 64 characters
    for char in unsanitized_session_name:
        if len(sanitized_session_name) == max_length:
            break
        if valid_characters_re.match(char):
            sanitized_session_name += char
    return sanitized_session_name



import asyncio
import copy
import ssl
import sys
from datetime import datetime, timedelta
from typing import Any, Dict, Optional

import bleach
import boto3
import requests as requests_sync
import tenacity
import ujson as json
from asgiref.sync import sync_to_async
from botocore.exceptions import ClientError
from cloudaux.aws.iam import (
    get_role_inline_policies,
    get_role_managed_policies,
    get_user_inline_policies,
    get_user_managed_policies,
    list_role_tags,
)
from cloudaux.aws.sts import boto3_cached_conn
from retrying import retry
from tornado.httpclient import AsyncHTTPClient
from tornado.httputil import url_concat

from consoleme.config import config
from consoleme.exceptions.exceptions import UserRoleNotAssumableYet
from consoleme.lib.aws import raise_if_background_check_required_and_no_background_check
from consoleme.lib.dynamo import IAMRoleDynamoHandler
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.policies import send_communications_policy_change_request_v2
from consoleme.lib.redis import RedisHandler

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()

log = config.get_logger(__name__)


class Aws:
    """The AWS class handles interactions with AWS."""

    def __init__(self) -> None:
        self.red = RedisHandler().redis_sync()
        self.redis_key = config.get("aws.iamroles_redis_key", "IAM_ROLE_CACHE")
        self.dynamo = IAMRoleDynamoHandler()

    @retry(
        stop_max_attempt_number=3,
        wait_exponential_multiplier=1000,
        wait_exponential_max=1000,
    )
    def _add_role_to_redis(self, role_entry: dict):
        """Add the role to redis with a retry.

        :param role_entry:
        :return:
        """
        self.red.hset(self.redis_key, role_entry["arn"], json.dumps(role_entry))

    @retry(
        stop_max_attempt_number=3,
        wait_exponential_multiplier=1000,
        wait_exponential_max=1000,
    )
    def _fetch_role_from_redis(self, role_arn: str):
        """Fetch the role from redis with a retry.

        :param role_arn:
        :return:
        """
        return self.red.hget(self.redis_key, role_arn)

    async def _cloudaux_to_aws(self, principal):
        """Convert the cloudaux get_role/get_user into the get_account_authorization_details equivalent."""
        # Pop out the fields that are not required:
        # Arn and RoleName/UserName will be popped off later:
        unrequired_fields = ["_version", "MaxSessionDuration"]
        principal_type = principal["Arn"].split(":")[-1].split("/")[0]
        for uf in unrequired_fields:
            principal.pop(uf, None)

        # Fix the Managed Policies:
        principal["AttachedManagedPolicies"] = list(
            map(
                lambda x: {"PolicyName": x["name"], "PolicyArn": x["arn"]},
                principal.get("ManagedPolicies", []),
            )
        )
        principal.pop("ManagedPolicies", None)

        # Fix the tags:
        if isinstance(principal.get("Tags", {}), dict):
            principal["Tags"] = list(
                map(
                    lambda key: {"Key": key, "Value": principal["Tags"][key]},
                    principal.get("Tags", {}),
                )
            )

        # Note: the instance profile list is verbose -- not transforming it (outside of renaming the field)!
        principal["InstanceProfileList"] = principal.pop("InstanceProfiles", [])

        # Inline Policies:
        if principal_type == "role":

            principal["RolePolicyList"] = list(
                map(
                    lambda name: {
                        "PolicyName": name,
                        "PolicyDocument": principal["InlinePolicies"][name],
                    },
                    principal.get("InlinePolicies", {}),
                )
            )
        else:
            principal["UserPolicyList"] = copy.deepcopy(
                principal.pop("InlinePolicies", [])
            )
        principal.pop("InlinePolicies", None)

        return principal

    @staticmethod
    def _get_iam_user_sync(account_id, user_name, conn) -> Optional[Dict[str, Any]]:
        client = boto3_cached_conn(
            "iam",
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            read_only=True,
            retry_max_attempts=2,
            client_kwargs=config.get("boto3.client_kwargs", {}),
        )
        user = client.get_user(UserName=user_name)["User"]
        user["ManagedPolicies"] = get_user_managed_policies(
            {"UserName": user_name}, **conn
        )
        user["InlinePolicies"] = get_user_inline_policies(
            {"UserName": user_name}, **conn
        )
        user["Tags"] = client.list_user_tags(UserName=user_name)
        user["Groups"] = client.list_groups_for_user(UserName=user_name)
        return user

    @staticmethod
    async def _get_iam_user_async(
        account_id, user_name, conn
    ) -> Optional[Dict[str, Any]]:
        tasks = []
        client = await sync_to_async(boto3_cached_conn)(
            "iam",
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            read_only=True,
            retry_max_attempts=2,
            client_kwargs=config.get("boto3.client_kwargs", {}),
        )
        user_details = asyncio.ensure_future(
            sync_to_async(client.get_user)(UserName=user_name)
        )
        tasks.append(user_details)

        all_tasks = [
            get_user_managed_policies,
            get_user_inline_policies,
        ]

        for t in all_tasks:
            tasks.append(
                asyncio.ensure_future(sync_to_async(t)({"UserName": user_name}, **conn))
            )

        user_tag_details = asyncio.ensure_future(
            sync_to_async(client.list_user_tags)(UserName=user_name)
        )
        tasks.append(user_tag_details)

        user_group_details = asyncio.ensure_future(
            sync_to_async(client.list_groups_for_user)(UserName=user_name)
        )
        tasks.append(user_group_details)

        responses = asyncio.gather(*tasks)
        async_task_result = await responses
        user = async_task_result[0]["User"]
        user["ManagedPolicies"] = async_task_result[1]
        inline_policies = []
        for name, policy in async_task_result[2].items():
            inline_policies.append({"PolicyName": name, "PolicyDocument": policy})
        user["InlinePolicies"] = inline_policies
        user["Tags"] = async_task_result[3].get("Tags", [])
        user["Groups"] = async_task_result[4].get("Groups", [])
        return user

    @staticmethod
    def _get_iam_role_sync(account_id, role_name, conn) -> Optional[Dict[str, Any]]:
        client = boto3_cached_conn(
            "iam",
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            read_only=True,
            retry_max_attempts=2,
            client_kwargs=config.get("boto3.client_kwargs", {}),
        )
        role = client.get_role(RoleName=role_name)["Role"]
        role["ManagedPolicies"] = get_role_managed_policies(
            {"RoleName": role_name}, **conn
        )
        role["InlinePolicies"] = get_role_inline_policies(
            {"RoleName": role_name}, **conn
        )
        role["Tags"] = list_role_tags({"RoleName": role_name}, **conn)
        return role

    @staticmethod
    async def _get_iam_role_async(
        account_id, role_name, conn
    ) -> Optional[Dict[str, Any]]:
        tasks = []
        client = await sync_to_async(boto3_cached_conn)(
            "iam",
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            read_only=True,
            retry_max_attempts=2,
            client_kwargs=config.get("boto3.client_kwargs", {}),
        )
        role_details = asyncio.ensure_future(
            sync_to_async(client.get_role)(RoleName=role_name)
        )
        tasks.append(role_details)

        all_tasks = [
            get_role_managed_policies,
            get_role_inline_policies,
            list_role_tags,
        ]

        for t in all_tasks:
            tasks.append(
                asyncio.ensure_future(sync_to_async(t)({"RoleName": role_name}, **conn))
            )

        responses = asyncio.gather(*tasks)
        async_task_result = await responses
        role = async_task_result[0]["Role"]
        role["ManagedPolicies"] = async_task_result[1]
        role["InlinePolicies"] = async_task_result[2]
        role["Tags"] = async_task_result[3]
        return role

    async def fetch_iam_user(
        self,
        account_id: str,
        user_arn: str,
        run_sync=False,
    ) -> Optional[Dict[str, Any]]:
        """Fetch the IAM User from AWS in threadpool if run_sync=False, otherwise synchronously.

        :param account_id:
        :param user_arn:
        :return:
        """
        log_data: dict = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user_arn": user_arn,
            "account_id": account_id,
        }

        try:
            user_name = user_arn.split("/")[-1]
            conn = {
                "account_number": account_id,
                "assume_role": config.get("policies.role_name"),
                "region": config.region,
                "client_kwargs": config.get("boto3.client_kwargs", {}),
            }
            if run_sync:
                user = self._get_iam_user_sync(account_id, user_name, conn)
            else:
                user = await self._get_iam_user_async(account_id, user_name, conn)

        except ClientError as ce:
            if ce.response["Error"]["Code"] == "NoSuchEntity":
                # The user does not exist:
                log_data["message"] = "User does not exist in AWS."
                log.error(log_data)
                stats.count(
                    "aws.fetch_iam_user.missing_in_aws",
                    tags={"account_id": account_id, "user_arn": user_arn},
                )
                return None

            else:
                log_data["message"] = f"Some other error: {ce.response}"
                log.error(log_data)
                stats.count(
                    "aws.fetch_iam_user.aws_connection_problem",
                    tags={"account_id": account_id, "user_arn": user_arn},
                )
                raise
        await self._cloudaux_to_aws(user)
        return user

    async def fetch_iam_role(
        self,
        account_id: str,
        role_arn: str,
        force_refresh: bool = False,
        run_sync=False,
    ) -> Optional[Dict[str, Any]]:
        """Fetch the IAM Role template from Redis and/or Dynamo.

        :param account_id:
        :param role_arn:
        :return:
        """
        log_data: dict = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "role_arn": role_arn,
            "account_id": account_id,
            "force_refresh": force_refresh,
        }

        result: dict = {}

        if not force_refresh:
            # First check redis:
            result: str = await sync_to_async(self._fetch_role_from_redis)(role_arn)

            if result:
                result: dict = json.loads(result)

                # If this item is less than an hour old, then return it from Redis.
                if result["ttl"] > int(
                    (datetime.utcnow() - timedelta(hours=1)).timestamp()
                ):
                    log_data["message"] = "Role not in Redis -- fetching from DDB."
                    log.debug(log_data)
                    stats.count(
                        "aws.fetch_iam_role.in_redis",
                        tags={"account_id": account_id, "role_arn": role_arn},
                    )
                    result["policy"] = json.loads(result["policy"])
                    return result

            # If not in Redis or it's older than an hour, proceed to DynamoDB:
            result = await sync_to_async(self.dynamo.fetch_iam_role)(
                role_arn, account_id
            )

        # If it's NOT in dynamo, or if we're forcing a refresh, we need to reach out to AWS and fetch:
        if force_refresh or not result.get("Item"):
            if force_refresh:
                log_data["message"] = "Force refresh is enabled. Going out to AWS."
                stats.count(
                    "aws.fetch_iam_role.force_refresh",
                    tags={"account_id": account_id, "role_arn": role_arn},
                )
            else:
                log_data["message"] = "Role is missing in DDB. Going out to AWS."
                stats.count(
                    "aws.fetch_iam_role.missing_dynamo",
                    tags={"account_id": account_id, "role_arn": role_arn},
                )
            log.debug(log_data)
            try:
                role_name = role_arn.split("/")[-1]
                conn = {
                    "account_number": account_id,
                    "assume_role": config.get("policies.role_name"),
                    "region": config.region,
                    "client_kwargs": config.get("boto3.client_kwargs", {}),
                }
                if run_sync:
                    role = self._get_iam_role_sync(account_id, role_name, conn)
                else:
                    role = await self._get_iam_role_async(account_id, role_name, conn)

            except ClientError as ce:
                if ce.response["Error"]["Code"] == "NoSuchEntity":
                    # The role does not exist:
                    log_data["message"] = "Role does not exist in AWS."
                    log.error(log_data)
                    stats.count(
                        "aws.fetch_iam_role.missing_in_aws",
                        tags={"account_id": account_id, "role_arn": role_arn},
                    )
                    return None

                else:
                    log_data["message"] = f"Some other error: {ce.response}"
                    log.error(log_data)
                    stats.count(
                        "aws.fetch_iam_role.aws_connection_problem",
                        tags={"account_id": account_id, "role_arn": role_arn},
                    )
                    raise

            # Format the role for DynamoDB and Redis:
            await self._cloudaux_to_aws(role)
            result = {
                "arn": role.get("Arn"),
                "name": role.pop("RoleName"),
                "resourceId": role.pop("RoleId"),
                "accountId": account_id,
                "ttl": int((datetime.utcnow() + timedelta(hours=36)).timestamp()),
                "policy": self.dynamo.convert_iam_resource_to_json(role),
                "permissions_boundary": role.get("PermissionsBoundary", {}),
                "templated": self.red.hget(
                    config.get("templated_roles.redis_key", "TEMPLATED_ROLES_v2"),
                    role.get("Arn").lower(),
                ),
            }

            # Sync with DDB:
            await sync_to_async(self.dynamo.sync_iam_role_for_account)(result)
            log_data["message"] = "Role fetched from AWS, and synced with DDB."
            stats.count(
                "aws.fetch_iam_role.fetched_from_aws",
                tags={"account_id": account_id, "role_arn": role_arn},
            )

        else:
            log_data["message"] = "Role fetched from DDB."
            stats.count(
                "aws.fetch_iam_role.in_dynamo",
                tags={"account_id": account_id, "role_arn": role_arn},
            )

            # Fix the TTL:
            result["Item"]["ttl"] = int(result["Item"]["ttl"])
            result = result["Item"]

        # Update the redis cache:
        stats.count(
            "aws.fetch_iam_role.in_dynamo",
            tags={"account_id": account_id, "role_arn": role_arn},
        )
        await sync_to_async(self._add_role_to_redis)(result)

        log_data["message"] += " Updated Redis."
        log.debug(log_data)

        result["policy"] = json.loads(result["policy"])
        return result

    async def call_user_lambda(
        self, role: str, user_email: str, account_id: str, user_role_name: str = "user"
    ) -> str:
        """Call out to the lambda function to provision the per-user role for the account."""
        raise NotImplementedError("This feature isn't enabled in ConsoleMe OSS")

    @tenacity.retry(
        wait=tenacity.wait_fixed(2),
        stop=tenacity.stop_after_attempt(10),
        retry=tenacity.retry_if_exception_type(UserRoleNotAssumableYet),
    )
    async def get_credentials(
        self,
        user: str,
        role: str,
        enforce_ip_restrictions: bool = True,
        user_role: bool = False,
        account_id: str = None,
        custom_ip_restrictions: list = None,
    ) -> dict:
        """Get Credentials will return the list of temporary credentials from AWS."""
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": user,
            "role": role,
            "enforce_ip_restrictions": enforce_ip_restrictions,
            "custom_ip_restrictions": custom_ip_restrictions,
            "message": "Generating credentials",
        }
        session = boto3.Session()
        client = session.client(
            "sts",
            region_name=config.region,
            endpoint_url=f"https://sts.{config.region}.amazonaws.com",
        )

        ip_restrictions = config.get("aws.ip_restrictions")
        stats.count("aws.get_credentials", tags={"role": role, "user": user})

        # If this is a dynamic request, then we need to fetch the role details, call out to the lambda
        # wait for it to complete, assume the role, and then return the assumed credentials back.
        if user_role:
            stats.count("aws.call_user_lambda", tags={"role": role, "user": user})
            try:
                role = await self.call_user_lambda(role, user, account_id)
            except Exception as e:
                raise e

        await raise_if_background_check_required_and_no_background_check(role, user)

        try:
            if enforce_ip_restrictions and ip_restrictions:
                policy = json.dumps(
                    dict(
                        Version="2012-10-17",
                        Statement=[
                            dict(
                                Effect="Deny",
                                Action="*",
                                Resource="*",
                                Condition=dict(
                                    NotIpAddress={"aws:SourceIP": ip_restrictions},
                                    Null={
                                        "aws:ViaAWSService": "true",
                                        "aws:PrincipalTag/AWSServiceTrust": "true",
                                    },
                                    StringNotLike={
                                        "aws:PrincipalArn": [
                                            "arn:aws:iam::*:role/aws:*"
                                        ]
                                    },
                                ),
                            ),
                            dict(Effect="Allow", Action="*", Resource="*"),
                        ],
                    )
                )

                credentials = await sync_to_async(client.assume_role)(
                    RoleArn=role,
                    RoleSessionName=user.lower(),
                    Policy=policy,
                    DurationSeconds=config.get("aws.session_duration", 3600),
                )
                credentials["Credentials"]["Expiration"] = int(
                    credentials["Credentials"]["Expiration"].timestamp()
                )
                log.debug(
                    {
                        **log_data,
                        "access_key_id": credentials["Credentials"]["AccessKeyId"],
                    }
                )
                return credentials
            if custom_ip_restrictions:
                policy = json.dumps(
                    dict(
                        Version="2012-10-17",
                        Statement=[
                            dict(
                                Effect="Deny",
                                Action="*",
                                Resource="*",
                                Condition=dict(
                                    NotIpAddress={
                                        "aws:SourceIP": custom_ip_restrictions
                                    },
                                    Null={
                                        "aws:ViaAWSService": "true",
                                        "aws:PrincipalTag/AWSServiceTrust": "true",
                                    },
                                    StringNotLike={
                                        "aws:PrincipalArn": [
                                            "arn:aws:iam::*:role/aws:*"
                                        ]
                                    },
                                ),
                            ),
                            dict(Effect="Allow", Action="*", Resource="*"),
                        ],
                    )
                )

                credentials = await sync_to_async(client.assume_role)(
                    RoleArn=role,
                    RoleSessionName=user.lower(),
                    Policy=policy,
                    DurationSeconds=config.get("aws.session_duration", 3600),
                )
                credentials["Credentials"]["Expiration"] = int(
                    credentials["Credentials"]["Expiration"].timestamp()
                )
                log.debug(
                    {
                        **log_data,
                        "access_key_id": credentials["Credentials"]["AccessKeyId"],
                    }
                )
                return credentials

            credentials = await sync_to_async(client.assume_role)(
                RoleArn=role,
                RoleSessionName=user.lower(),
                DurationSeconds=config.get("aws.session_duration", 3600),
            )
            credentials["Credentials"]["Expiration"] = int(
                credentials["Credentials"]["Expiration"].timestamp()
            )
            log.debug(
                {**log_data, "access_key_id": credentials["Credentials"]["AccessKeyId"]}
            )
            return credentials
        except ClientError as e:
            # TODO(ccastrapel): Determine if user role was really just created, or if this is an older role.
            if user_role:
                raise UserRoleNotAssumableYet(e.response["Error"])
            raise

    async def generate_url(
        self,
        user: str,
        role: str,
        region: str = "us-east-1",
        user_role: bool = False,
        account_id: str = None,
    ) -> str:
        """Generate URL will get temporary credentials and craft a URL with those credentials."""
        function = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        log_data = {
            "function": function,
            "user": user,
            "role": role,
            "message": "Generating authenticated AWS console URL",
        }
        log.debug(log_data)
        credentials = await self.get_credentials(
            user,
            role,
            user_role=user_role,
            account_id=account_id,
            enforce_ip_restrictions=False,
        )

        credentials_d = {
            "sessionId": credentials.get("Credentials", {}).get("AccessKeyId"),
            "sessionKey": credentials.get("Credentials", {}).get("SecretAccessKey"),
            "sessionToken": credentials.get("Credentials", {}).get("SessionToken"),
        }

        req_params = {
            "Action": "getSigninToken",
            "Session": bleach.clean(json.dumps(credentials_d)),
            "DurationSeconds": config.get("aws.session_duration", 3600),
        }

        http_client = AsyncHTTPClient(force_instance=True)

        url_with_params: str = url_concat(
            config.get(
                "aws.federation_url", "https://signin.aws.amazon.com/federation"
            ),
            req_params,
        )
        r = await http_client.fetch(url_with_params, ssl_options=ssl.SSLContext())
        token = json.loads(r.body)

        login_req_params = {
            "Action": "login",
            "Issuer": config.get("aws.issuer"),
            "Destination": (
                "{}".format(
                    config.get(
                        "aws.console_url", "https://{}.console.aws.amazon.com"
                    ).format(region)
                )
            ),
            "SigninToken": bleach.clean(token.get("SigninToken")),
            "SessionDuration": config.get("aws.session_duration", 3600),
        }

        r2 = requests_sync.Request(
            "GET",
            config.get(
                "aws.federation_url", "https://signin.aws.amazon.com/federation"
            ),
            params=login_req_params,
        )
        url = r2.prepare().url
        return url

    async def sns_publisher_group_requests(
        self, user, group, justification, request_id, bg_check_passed
    ):
        raise NotImplementedError()

    async def sns_publish_policy_requests(self, request, request_uri):
        raise NotImplementedError()

    async def send_communications_policy_change_request(self, request, send_sns=False):
        """
        Optionally send a notification when there's a new policy change request

        :param request:
        :param send_sns:
        :return:
        """
        log_data: dict = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Function is not configured.",
        }
        log.warning(log_data)
        return

    async def send_communications_new_policy_request(
        self, extended_request, admin_approved, approval_probe_approved
    ):
        """
        Optionally send a notification when there's a new policy change request

        :param approval_probe_approved:
        :param admin_approved:
        :param extended_request:
        :return:
        """
        await send_communications_policy_change_request_v2(extended_request)
        return

    @staticmethod
    def handle_detected_role(role):
        pass

    async def should_auto_approve_policy_v2(self, extended_request, user, user_groups):
        return {"approved": False}


def init():
    """Initialize the AWS plugin."""
    return Aws()



import sys
import traceback

import ujson as json
from asgiref.sync import sync_to_async
from botocore.exceptions import ClientError
from cloudaux.aws.sts import boto3_cached_conn

from consoleme.config import config
from consoleme.lib.aws import sanitize_session_name
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.role_updater.schemas import RoleUpdaterRequest

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


async def update_role(event):
    log_data = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "event": event,
        "message": "Working on event",
    }
    log.debug(log_data)

    if not isinstance(event, list):
        raise Exception("The passed event must be a list.")

    # Let's normalize all of the policies to JSON if they are not already
    for d in event:
        for i in d.get("inline_policies", []):
            if i.get("policy_document") and isinstance(i.get("policy_document"), dict):
                i["policy_document"] = json.dumps(
                    i["policy_document"], escape_forward_slashes=False
                )

        if d.get("assume_role_policy_document", {}):
            if isinstance(
                d.get("assume_role_policy_document", {}).get(
                    "assume_role_policy_document"
                ),
                dict,
            ):
                d["assume_role_policy_document"][
                    "assume_role_policy_document"
                ] = json.dumps(
                    d["assume_role_policy_document"]["assume_role_policy_document"],
                    escape_forward_slashes=False,
                )

    bad_validation = RoleUpdaterRequest().validate(event, many=True)
    if bad_validation:
        log_data["error"] = bad_validation
        log.error(log_data)
        return {"error_msg": "invalid schema passed", "detail_error": bad_validation}

    event = RoleUpdaterRequest().load(event, many=True)

    result = {"success": False}

    for d in event:
        arn = d["arn"]
        aws_session_name = sanitize_session_name("roleupdater-" + d["requester"])
        account_number = await parse_account_id_from_arn(arn)
        role_name = await parse_role_name_from_arn(arn)
        # TODO: Make configurable
        client = boto3_cached_conn(
            "iam",
            account_number=account_number,
            assume_role=config.get("policies.role_name", "ConsoleMe"),
            session_name=aws_session_name,
            retry_max_attempts=2,
            client_kwargs=config.get("boto3.client_kwargs", {}),
        )
        inline_policies = d.get("inline_policies", [])
        managed_policies = d.get("managed_policies", [])
        assume_role_doc = d.get("assume_role_policy_document", {})
        tags = d.get("tags", [])

        if (
            not inline_policies
            and not managed_policies
            and not assume_role_doc
            and not tags
        ):
            result["message"] = f"Invalid request. No response taken on event: {event}"
            return result

        try:
            for policy in inline_policies:
                await update_inline_policy(client, role_name, policy)

            for policy in managed_policies:
                await update_managed_policy(client, role_name, policy)

            if assume_role_doc:
                await update_assume_role_document(client, role_name, assume_role_doc)

            for tag in tags:
                await update_tags(client, role_name, tag)
        except ClientError as ce:
            result["message"] = ce.response["Error"]
            result["Traceback"] = traceback.format_exc()
            return result
        result["success"] = True
        return result


async def parse_account_id_from_arn(arn):
    return arn.split(":")[4]


async def parse_role_name_from_arn(arn):
    return arn.split("/")[-1]


async def update_inline_policy(client, role_name, policy):
    log.debug(
        {"message": "Updating inline policy", "role_name": role_name, "policy": policy}
    )
    if policy.get("action") == "attach":
        response = await sync_to_async(client.put_role_policy)(
            RoleName=role_name,
            PolicyName=policy["policy_name"],
            PolicyDocument=policy["policy_document"],
        )
    elif policy.get("action") == "detach":
        response = await sync_to_async(client.delete_role_policy)(
            RoleName=role_name, PolicyName=policy["policy_name"]
        )
    else:
        raise Exception("Unable to update managed policy")
    return response


async def update_managed_policy(client, role_name, policy):
    log.debug(
        {"message": "Updating managed policy", "role_name": role_name, "policy": policy}
    )
    if policy.get("action") == "attach":
        response = await sync_to_async(client.attach_role_policy)(
            PolicyArn=policy["arn"], RoleName=role_name
        )
    elif policy.get("action") == "detach":
        response = await sync_to_async(client.detach_role_policy)(
            PolicyArn=policy["arn"], RoleName=role_name
        )
    else:
        raise Exception("Unable to update managed policy.")
    return response


async def update_assume_role_document(client, role_name, assume_role_doc):
    log.debug(
        {
            "message": "Updating assume role doc",
            "role_name": role_name,
            "assume_role_doc": assume_role_doc,
        }
    )
    response = None
    if assume_role_doc.get("action", "") in ["create", "update"]:
        response = await sync_to_async(client.update_assume_role_policy)(
            RoleName=role_name,
            PolicyDocument=assume_role_doc["assume_role_policy_document"],
        )
    return response
    # Log or report result?


async def update_tags(client, role_name, tag):
    log.debug({"message": "Updating tag", "role_name": role_name, "tag": tag})
    if tag.get("action") == "add":
        response = await sync_to_async(client.tag_role)(
            RoleName=role_name, Tags=[{"Key": tag["key"], "Value": tag["value"]}]
        )
    elif tag.get("action") == "remove":
        response = await sync_to_async(client.untag_role)(
            RoleName=role_name, TagKeys=[tag["key"]]
        )
    else:
        raise Exception("Unable to update tags.")
    return response



from typing import Dict, Optional, Union


class Metric:
    def count(self, metric_name, tags=None):
        raise NotImplementedError

    def gauge(self, metric_name, metric_value, tags=None):
        raise NotImplementedError

    def timer(
        self,
        metric_name: str,
        tags: Optional[Union[Dict[str, Union[str, bool]], Dict[str, str]]] = None,
    ) -> None:
        raise NotImplementedError



import sys
import uuid
from datetime import datetime, timedelta

import pytz
import tornado.escape
import ujson as json
from asgiref.sync import async_to_sync

from consoleme.config import config
from consoleme.exceptions.exceptions import MissingConfigurationValue
from consoleme.handlers.base import BaseHandler, TornadoRequestHandler
from consoleme.lib.challenge import delete_expired_challenges, retrieve_user_challenge
from consoleme.lib.jwt import generate_jwt_token
from consoleme.lib.redis import RedisHandler

log = config.get_logger()
red = async_to_sync(RedisHandler().redis)()


class ChallengeGeneratorHandler(TornadoRequestHandler):
    """
    Challenge URLs are an alternative to mutual TLS for authenticating CLI clients of ConsoleMe.

    If Challenge Token auth is enabled, this will generate time-sensitive challenge token urls that end-users
    will be required to authenticate to. One authentication is verified, clients will be able to retrieve a
    signed jwt that clients will be able to pass to ConsoleMe for authn/authz.

    The ChallengeUrlGenerator endpoint must be unauthenticated because the CLI client will be requesting URLs
    """

    def get_request_ip(self):
        trusted_remote_ip_header = config.get("auth.remote_ip.trusted_remote_ip_header")
        if trusted_remote_ip_header:
            return self.request.headers[trusted_remote_ip_header].split(",")[0]
        return self.request.remote_ip

    async def get(self, user):
        if not config.get("challenge_url.enabled", False):
            raise MissingConfigurationValue(
                "Challenge URL Authentication is not enabled in ConsoleMe's configuration"
            )
        ip = self.get_request_ip()

        token = str(uuid.uuid4())
        entry = {
            "ttl": int(
                (
                    datetime.utcnow().replace(tzinfo=pytz.UTC) + timedelta(minutes=2)
                ).timestamp()
            ),
            "ip": ip,
            "status": "pending",
            "user": user,
        }
        red.hset(
            config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP"),
            token,
            json.dumps(entry),
        )

        challenge_url = "{url}/challenge_validator/{token}".format(
            url=config.get("url"), token=token
        )
        polling_url = "{url}/noauth/v1/challenge_poller/{token}".format(
            url=config.get("url"), token=token
        )
        self.write({"challenge_url": challenge_url, "polling_url": polling_url})

        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "challenge_url": challenge_url,
            "polling_url": polling_url,
            "message": "Incoming request",
            "ip": ip,
            "user": user,
        }
        log.debug(log_data)


class ChallengeValidatorHandler(BaseHandler):
    """

    This is the challenge authentication endpoint.
    Once the user has authenticated successfully, we validate their information and mark the challenge as successful.

    """

    async def get(self, requested_challenge_token):
        if not config.get("challenge_url.enabled", False):
            raise MissingConfigurationValue(
                "Challenge URL Authentication is not enabled in ConsoleMe's configuration"
            )
        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "requested_challenge_token": requested_challenge_token,
            "message": "Incoming request",
            "ip": self.ip,
        }
        log.debug(log_data)

        all_challenges = red.hgetall(
            config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP")
        )
        if not all_challenges:
            message = (
                "Unable to find a matching challenge URL. This usually means that it has expired. "
                "Please try requesting a new challenge URL."
            )
            self.write({"message": message})
            return

        await delete_expired_challenges(all_challenges)

        valid_user_challenge = await retrieve_user_challenge(
            self, requested_challenge_token, log_data
        )
        if not valid_user_challenge:
            return

        if valid_user_challenge.get("visited"):
            message = (
                "This unique challenge URL has already been viewed. "
                "Please try requesting a new challenge URL."
            )
            self.write({"message": message})
            return

        request_ip = self.get_request_ip()

        # By default, the challenge URL requester IP must match the URL the challenge was created with. In some cases
        # (i.e. IPv4 vs IPv6), the challenge may have been created with an IPv4 address, and the authenticated browser
        # verification request may originate from an IPv6 one, or visa versa, in which case this configuration may
        # need to be explicitly set to False.
        if config.get(
            "challenge_url.request_ip_must_match_challenge_creation_ip", True
        ):
            if request_ip != valid_user_challenge.get("ip"):
                log.error(
                    {
                        **log_data,
                        "request_ip": request_ip,
                        "challenge_ip": valid_user_challenge.get("ip"),
                        "message": "Request IP doesn't match challenge IP",
                    }
                )
                self.write(
                    {
                        "message": (
                            "Your originating IP doesn't match the IP the challenge was created with. "
                            "If you are developing locally, this is probably because your CLI (Weep) made an IPv6 "
                            "request, and your web browser made an IPv4 request. Or visa-versa. If this is the case, "
                            "set the local configuration for "
                            "**challenge_url.request_ip_must_match_challenge_creation_ip** to **false**."
                        )
                    }
                )
                return

        valid_user_challenge["visited"] = True
        valid_user_challenge["nonce"] = str(uuid.uuid4())
        red.hset(
            config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP"),
            requested_challenge_token,
            json.dumps(valid_user_challenge),
        )

        request_ip = valid_user_challenge["ip"]
        request_user = valid_user_challenge["user"]
        message = (
            f"A user at **{request_ip}** has requested ConsoleMe credentials for **{request_user}**.\n\n"
            f"You must approve this request for credentials to be provided. "
            f"You will not be able to refresh or revisit this page after closing it.\n\n"
            f"If you did not create this request, please report it to your security team."
        )

        self.write(
            {
                "message": message,
                "nonce": valid_user_challenge["nonce"],
                "show_approve_button": True,
            }
        )

    async def post(self, requested_challenge_token):
        if not config.get("challenge_url.enabled", False):
            raise MissingConfigurationValue(
                "Challenge URL Authentication is not enabled in ConsoleMe's configuration"
            )
        data = tornado.escape.json_decode(self.request.body)

        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "requested_challenge_token": requested_challenge_token,
            "message": "Incoming request",
            "ip": self.ip,
        }
        log.debug(log_data)

        all_challenges = red.hgetall(
            config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP")
        )
        if not all_challenges:
            message = (
                "Unable to find a matching challenge URL. This usually means that it has expired. "
                "Please try requesting a new challenge URL."
            )
            self.write({"message": message})
            return

        await delete_expired_challenges(all_challenges)

        valid_user_challenge = await retrieve_user_challenge(
            self, requested_challenge_token, log_data
        )
        if not valid_user_challenge:
            message = (
                "Unable to find a matching challenge URL. This usually means that it has expired. "
                "Please try requesting a new challenge URL."
            )
            self.write({"message": message})
            return

        if data.get("nonce") != valid_user_challenge["nonce"]:
            message = "Unable to validate challenge URL. The Nonce you've submitted is invalid."
            log.error({**log_data, "message": message})
            self.write({"message": message})
            return

        request_ip = self.get_request_ip()

        # By default, the challenge URL requester IP must match the URL the challenge was created with. In some cases
        # (i.e. IPv4 vs IPv6), the challenge may have been created with an IPv4 address, and the authenticated browser
        # verification request may originate from an IPv6 one, or visa versa, in which case this configuration may
        # need to be explicitly set to False.
        if config.get(
            "challenge_url.request_ip_must_match_challenge_creation_ip", True
        ):
            if request_ip != valid_user_challenge.get("ip"):
                log.error(
                    {
                        **log_data,
                        "request_ip": request_ip,
                        "challenge_ip": valid_user_challenge.get("ip"),
                        "message": "Request IP doesn't match challenge IP",
                    }
                )
                self.write(
                    {
                        "message": "Your originating IP doesn't match the IP the challenge was created with."
                    }
                )
                return

        valid_user_challenge["status"] = "success"
        valid_user_challenge["user"] = self.user
        valid_user_challenge["groups"] = self.groups
        red.hset(
            config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP"),
            requested_challenge_token,
            json.dumps(valid_user_challenge),
        )
        message = "You've successfully authenticated to ConsoleMe and may now close this page."
        self.write({"message": message})


class ChallengePollerHandler(TornadoRequestHandler):
    """
    This endpoint is an unauthenticated endpoint that the client uses to poll for successful challenge completion.
    If the challenge has been completed successfully, and the IP of the endpoint matches the IP used to generate the
    challenge URL, we return a signed jwt. It is expected that the client will poll this endpoint continuously until
    the challenge url has been validated by a client, or until it has expired.
    """

    async def get(self, requested_challenge_token):
        if not config.get("challenge_url.enabled", False):
            raise MissingConfigurationValue(
                "Challenge URL Authentication is not enabled in ConsoleMe's configuration"
            )
        challenge_j = red.hget(
            config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP"),
            requested_challenge_token,
        )
        if not challenge_j:
            self.write({"status": "unknown"})
            return
        challenge = json.loads(challenge_j)

        # Delete the token if it has expired
        current_time = int(datetime.utcnow().replace(tzinfo=pytz.UTC).timestamp())
        if challenge.get("ttl", 0) < current_time:
            red.hdel(
                config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP"),
                requested_challenge_token,
            )
            self.write({"status": "expired"})
            return

        ip = self.get_request_ip()

        if ip != challenge.get("ip"):
            self.write({"status": "unauthorized"})
            return

        # Generate a jwt if user authentication was successful
        if challenge.get("status") == "success":
            jwt_expiration = datetime.utcnow().replace(tzinfo=pytz.UTC) + timedelta(
                minutes=config.get("jwt.expiration_minutes", 60)
            )
            encoded_jwt = await generate_jwt_token(
                challenge.get("user"), challenge.get("groups"), exp=jwt_expiration
            )

            self.write(
                {
                    "status": challenge["status"],
                    "cookie_name": config.get("auth_cookie_name", "consoleme_auth"),
                    "expiration": int(jwt_expiration.timestamp()),
                    "encoded_jwt": encoded_jwt,
                    "user": challenge["user"],
                }
            )
            # Delete the token so that it cannot be re-used
            red.hdel(
                config.get("challenge_url.redis_key", "TOKEN_CHALLENGES_TEMP"),
                requested_challenge_token,
            )
            return
        self.write({"status": challenge.get("status")})
        return



import sys
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import boto3
import pytz
import ujson as json
from asgiref.sync import sync_to_async
from botocore.exceptions import ClientError
from cloudaux import sts_conn
from cloudaux.aws.decorators import rate_limited
from cloudaux.aws.sts import boto3_cached_conn

from consoleme.config import config
from consoleme.exceptions.exceptions import MissingConfigurationValue
from consoleme.lib.plugins import get_plugin_by_name

log = config.get_logger("consoleme")
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


async def is_object_older_than_seconds(
    key: str, older_than_seconds: int, bucket: Optional[str] = None, s3_client=None
) -> bool:
    """
    This function checks if an S3 object is older than the specified number of seconds. if the object doesn't
    exist, this function will return True.
    """
    if not bucket:
        bucket = config.get("consoleme_s3_bucket")
    if not bucket:
        raise MissingConfigurationValue(
            "`bucket` not defined, and we can't find the default bucket in "
            "the configuration key `consoleme_s3_bucket`."
        )
    now = datetime.utcnow().replace(tzinfo=pytz.utc)
    if not s3_client:
        s3_client = boto3.client("s3", **config.get("boto3.client_kwargs", {}))
    try:
        res = await sync_to_async(s3_client.head_object)(Bucket=bucket, Key=key)
    except ClientError as e:
        # If file is not found, we'll tell the user it's older than the specified time
        if e.response.get("Error", {}).get("Code") == "404":
            return True
        raise
    datetime_value = res["LastModified"]
    if (now - datetime_value).total_seconds() > older_than_seconds:
        return True
    return False


async def does_object_exist(bucket: str, key: str, s3_client=None) -> bool:
    """
    This function checks if an S3 object exists.
    """
    if not s3_client:
        s3_client = boto3.client("s3", **config.get("boto3.client_kwargs", {}))
    try:
        await sync_to_async(s3_client.head_object)(Bucket=bucket, Key=key)
    except ClientError as e:
        # If file is not found, we'll tell the user it's older than the specified time
        if e.response.get("Error", {}).get("Code") == "404":
            return False
        raise
    return True


@rate_limited()
@sts_conn("s3")
def put_object(client=None, **kwargs):
    """Create an S3 object -- calls wrapped with CloudAux."""
    return client.put_object(**kwargs)


def get_object(**kwargs):
    client = kwargs.get("client")
    assume_role = kwargs.get("assume_role")
    if not client:
        if assume_role:
            client = boto3_cached_conn(
                "s3",
                account_number=kwargs.get("account_number"),
                assume_role=assume_role,
                session_name=kwargs.get("session_name", "ConsoleMe"),
                region=kwargs.get("region", config.region),
                retry_max_attempts=2,
                client_kwargs=config.get("boto3.client_kwargs", {}),
            )
        else:
            client = boto3.client("s3", **config.get("boto3.client_kwargs", {}))
    return client.get_object(Bucket=kwargs.get("Bucket"), Key=kwargs.get("Key"))


async def get_object_async(**kwargs):
    """Get an S3 object Asynchronously"""
    return await sync_to_async(get_object)(**kwargs)


async def fetch_json_object_from_s3(
    bucket: str, object: str
) -> Union[Dict[str, Any], List[dict]]:
    """
    Fetch and load a JSON-formatted object in S3
    :param bucket: S3 bucket
    :param object: S3 Object
    :return: Dict
    """
    s3_object = await get_object_async(Bucket=bucket, Key=object, region=config.region)
    object_content = s3_object["Body"].read()
    data = json.loads(object_content)
    return data


def map_operation_to_api(operation, default):
    function: str = f"{__name__}.{sys._getframe().f_code.co_name}"
    operations_map = {
        "BATCH.DELETE.OBJECT": "s3:DeleteObject",
        "REST.COPY.OBJECT": "s3:PutObject",
        "REST.COPY.OBJECT_GET": "REST.COPY.OBJECT_GET",
        "REST.COPY.PART": "s3:PutObject",
        "REST.DELETE.BUCKET": "s3:DeleteBucket",
        "REST.DELETE.OBJECT": "s3:DeleteObject",
        "REST.DELETE.UPLOAD": "s3:DeleteObject",
        "REST.GET.ACCELERATE": "s3:GetAccelerateConfiguration",
        "REST.GET.ACL": "s3:GetObjectVersionAcl",
        "REST.GET.ANALYTICS": "s3:GetAnalyticsConfiguration",
        "REST.GET.BUCKET": "s3:GetBucket",
        "REST.GET.BUCKETPOLICY": "s3:GetBucketPolicy",
        "REST.GET.BUCKETVERSIONS": "s3:GetBucketVersioning",
        "REST.GET.CORS": "s3:GetBucketCORS",
        "REST.GET.ENCRYPTION": "s3:GetEncryptionConfiguration",
        "REST.GET.INTELLIGENT_TIERING": "REST.GET.INTELLIGENT_TIERING",
        "REST.GET.INVENTORY": "s3:GetInventoryConfiguration",
        "REST.GET.LIFECYCLE": "s3:GetLifecycleConfiguration",
        "REST.GET.LOCATION": "s3:GetBucketLocation",
        "REST.GET.LOGGING_STATUS": "s3:GetBucketLogging",
        "REST.GET.METRICS": "s3:GetMetricsConfiguration",
        "REST.GET.NOTIFICATION": "s3:GetBucketNotification",
        "REST.GET.OBJECT": "s3:GetObject",
        "REST.GET.OBJECT_LOCK_CONFIGURATION": "s3:GetObjectLockConfiguration",
        "REST.GET.OBJECT_TAGGING": "s3:GetObjectTagging",
        "REST.GET.OWNERSHIP_CONTROLS": "s3:GetBucketOwnershipControls",
        "REST.GET.POLICY_STATUS": "s3:GetBucketPolicyStatus",
        "REST.GET.PUBLIC_ACCESS_BLOCK": "s3:GetBucketPublicAccessBlock",
        "REST.GET.REPLICATION": "s3:GetReplicationConfiguration",
        "REST.GET.REQUEST_PAYMENT": "s3:GetBucketRequestPayment",
        "REST.GET.TAGGING": "s3:GetBucketTagging",
        "REST.GET.UPLOAD": "s3:GetObject",
        "REST.GET.UPLOADS": "s3:GetObject",
        "REST.GET.VERSIONING": "s3:GetObjectVersion",
        "REST.GET.WEBSITE": "s3:GetBucketWebsite",
        "REST.HEAD.BUCKET": "s3:ListBucket",
        "REST.HEAD.OBJECT": "s3:GetObject",
        "REST.COPY.PART_GET": "s3:GetObject",
        "REST.POST.BUCKET": "REST.POST.BUCKET",
        "REST.POST.MULTI_OBJECT_DELETE": "s3:DeleteObject",
        "REST.POST.RESTORE": "s3:RestoreObject",
        "REST.POST.UPLOAD": "s3:PutObject",
        "REST.POST.UPLOADS": "s3:PutObject",
        "REST.PUT.ACL": "s3:PutBucketAcl,s3:PutObjectAcl",
        "REST.PUT.BUCKET": "REST.PUT.BUCKET",
        "REST.PUT.BUCKETPOLICY": "s3:PutBucketPolicy",
        "REST.PUT.CORS": "s3:PutBucketCORS",
        "REST.PUT.LIFECYCLE": "s3:PutLifecycleConfiguration",
        "REST.PUT.LOGGING_STATUS": "REST.PUT.LOGGING_STATUS-NotBucketOwnerOrGrantee",
        "REST.PUT.NOTIFICATION": "s3:PutBucketNotification",
        "REST.PUT.OBJECT": "s3:PutObject",
        "REST.PUT.OBJECT_TAGGING": "s3:PutObjectTagging",
        "REST.PUT.PART": "s3:PutObject",
        "REST.PUT.PUBLIC_ACCESS_BLOCK": "s3:PutBucketPublicAccessBlock",
        "REST.PUT.REPLICATION": "s3:PutReplicationConfiguration",
        "REST.PUT.VERSIONING": "s3:putbucketversioning",
    }
    api_call = operations_map.get(operation)
    if api_call is None:
        stats.count(f"{function}.error")
        log.error(
            {
                "message": "S3 Operation Needs Mapping",
                "function": function,
                "query": operation,
            }
        )
        return default
    else:
        stats.count(f"{function}.success")
        return api_call



from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.plugins import get_plugin_by_name

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()

BASE_INLINE_POLICY = {"Statement": [{"Action": [], "Effect": "Allow", "Resource": []}]}


class GeneratePolicyHandler(BaseAPIV2Handler):
    """Handler for /api/v2/generate_policy

    Generates an AWS role / resource policy given a set of CRUD permissions
    """

    allowed_methods = ["GET", "POST"]

    async def get(self):
        self.write(BASE_INLINE_POLICY)

    async def post(self):
        """
        POST /api/v2/generate_policy

        Determine which user is requesting access to which resource, and the type of access based on their
        selections in self-service
        """
        self.write(BASE_INLINE_POLICY)
        tags = {"user": self.user}
        stats.count("RequestsHandler.post", tags=tags)
        log_data = {
            "function": "RequestsHandler.post",
            "user": self.user,
            "message": "Creating request",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        self.write_error(501, message="Create request")



import boto3
from botocore.exceptions import ClientError
from tenacity import Retrying, retry_if_exception_type, stop_after_attempt, wait_fixed

from consoleme.config import config

ddb = boto3.client(
    "dynamodb",
    endpoint_url=config.get(
        "dynamodb_server", config.get("boto3.client_kwargs.endpoint_url")
    ),
    region_name=config.region,
)

table_name = "consoleme_iamroles_global"
try:
    ddb.create_table(
        TableName=table_name,
        AttributeDefinitions=[
            {"AttributeName": "arn", "AttributeType": "S"},
            {"AttributeName": "accountId", "AttributeType": "S"},
        ],
        KeySchema=[
            {"AttributeName": "arn", "KeyType": "HASH"},
            {"AttributeName": "accountId", "KeyType": "RANGE"},
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 100, "WriteCapacityUnits": 100},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )

    try:
        for attempt in Retrying(
            stop=stop_after_attempt(3),
            wait=wait_fixed(3),
            retry=retry_if_exception_type(
                (
                    ddb.exceptions.ResourceNotFoundException,
                    ddb.exceptions.ResourceInUseException,
                )
            ),
        ):
            with attempt:
                ddb.update_time_to_live(
                    TableName=table_name,
                    TimeToLiveSpecification={"Enabled": True, "AttributeName": "ttl"},
                )
    except ClientError as e:
        if str(e) != (
            "An error occurred (ValidationException) when calling the UpdateTimeToLive operation: "
            "TimeToLive is already enabled"
        ):
            print(f"Unable to update TTL attribute on table {table_name}. Error: {e}.")

except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")

table_name = "consoleme_config_global"
try:
    ddb.create_table(
        TableName="consoleme_config_global",
        KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],  # Partition key
        AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
        ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )
except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")

table_name = "consoleme_policy_requests"
try:
    ddb.create_table(
        TableName=table_name,
        KeySchema=[{"AttributeName": "request_id", "KeyType": "HASH"}],  # Partition key
        AttributeDefinitions=[
            {"AttributeName": "request_id", "AttributeType": "S"},
            {"AttributeName": "arn", "AttributeType": "S"},
        ],
        GlobalSecondaryIndexes=[
            {
                "IndexName": "arn-request_id-index",
                "KeySchema": [{"AttributeName": "arn", "KeyType": "HASH"}],
                "Projection": {"ProjectionType": "ALL"},
                "ProvisionedThroughput": {
                    "ReadCapacityUnits": 123,
                    "WriteCapacityUnits": 123,
                },
            }
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )
except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")

table_name = "consoleme_resource_cache"
try:
    ddb.create_table(
        TableName=table_name,
        KeySchema=[
            {"AttributeName": "resourceId", "KeyType": "HASH"},
            {"AttributeName": "resourceType", "KeyType": "RANGE"},  # Sort key
        ],  # Partition key
        AttributeDefinitions=[
            {"AttributeName": "resourceId", "AttributeType": "S"},
            {"AttributeName": "resourceType", "AttributeType": "S"},
            {"AttributeName": "arn", "AttributeType": "S"},
        ],
        GlobalSecondaryIndexes=[
            {
                "IndexName": "arn-index",
                "KeySchema": [{"AttributeName": "arn", "KeyType": "HASH"}],
                "Projection": {"ProjectionType": "ALL"},
                "ProvisionedThroughput": {
                    "ReadCapacityUnits": 123,
                    "WriteCapacityUnits": 123,
                },
            }
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )
except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")

table_name = "consoleme_cloudtrail"
try:
    ddb.create_table(
        TableName=table_name,
        KeySchema=[
            {"AttributeName": "arn", "KeyType": "HASH"},  # Partition key
            {"AttributeName": "request_id", "KeyType": "RANGE"},  # Sort key
        ],
        AttributeDefinitions=[
            {"AttributeName": "arn", "AttributeType": "S"},
            {"AttributeName": "request_id", "AttributeType": "S"},
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )
    try:
        for attempt in Retrying(
            stop=stop_after_attempt(3),
            wait=wait_fixed(3),
            retry=retry_if_exception_type(
                (
                    ddb.exceptions.ResourceNotFoundException,
                    ddb.exceptions.ResourceInUseException,
                )
            ),
        ):
            with attempt:
                ddb.update_time_to_live(
                    TableName=table_name,
                    TimeToLiveSpecification={"Enabled": True, "AttributeName": "ttl"},
                )
    except ClientError as e:
        if str(e) != (
            "An error occurred (ValidationException) when calling the UpdateTimeToLive operation: "
            "TimeToLive is already enabled"
        ):
            print(f"Unable to update TTL attribute on table {table_name}. Error: {e}.")
except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")

try:
    table_name = "consoleme_users_global"
    ddb.create_table(
        TableName=table_name,
        KeySchema=[{"AttributeName": "username", "KeyType": "HASH"}],  # Partition key
        AttributeDefinitions=[{"AttributeName": "username", "AttributeType": "S"}],
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )
except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")


table_name = "consoleme_notifications"
try:
    ddb.create_table(
        TableName=table_name,
        AttributeDefinitions=[
            {"AttributeName": "predictable_id", "AttributeType": "S"},
        ],
        KeySchema=[
            {"AttributeName": "predictable_id", "KeyType": "HASH"},
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 100, "WriteCapacityUnits": 100},
        StreamSpecification={
            "StreamEnabled": True,
            "StreamViewType": "NEW_AND_OLD_IMAGES",
        },
    )

except ClientError as e:
    if str(e) != (
        "An error occurred (ResourceInUseException) when calling the CreateTable operation: "
        "Cannot create preexisting table"
    ):
        print(f"Unable to create table {table_name}. Error: {e}.")



"""
Helpers library for various actions
"""

import os
import subprocess
from pathlib import Path

from aws_cdk import aws_lambda as lambda_


def create_dependencies_layer(self, function_name: str) -> lambda_.LayerVersion:
    base_path = f"./resources/{function_name}"
    output_dir = f"{base_path}/build/{function_name}"
    if not os.environ.get("SKIP_PIP"):
        build_folder = Path(f"{base_path}/build")
        requirements_file = Path(f"{base_path}/requirements.txt")
        # Remove build folder from previous runs
        if build_folder.is_dir():
            subprocess.check_call(f"rm -rf {base_path}/build".split())
        # Remove requirements file from previous runs
        if requirements_file.is_file():
            subprocess.check_call(f"rm -rf {base_path}/requirements.txt".split())
        # Create requirements file using pipreqs
        subprocess.check_call(f"pipreqs {base_path}".split())
        requirements_file = Path(f"{base_path}/requirements.txt")
        if requirements_file.is_file():
            subprocess.check_call(
                f"pip install -r {base_path}/requirements.txt -t {output_dir}/python".split()
            )
    layer_id = f"{function_name}-dependencies"
    layer_code = lambda_.Code.from_asset(output_dir)
    return lambda_.LayerVersion(self, layer_id, code=layer_code)



"""
Database stack for running ConsoleMe on ECS
"""

from aws_cdk import aws_dynamodb as db
from aws_cdk import core as cdk


class DBStack(cdk.NestedStack):
    """
    Database stack for running ConsoleMe on ECS
    """

    def __init__(self, scope: cdk.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        # DynamoDB tables

        db.Table(
            self,
            "IAMRolesTable",
            table_name="consoleme_iamroles_global",
            partition_key=db.Attribute(name="arn", type=db.AttributeType.STRING),
            sort_key=db.Attribute(name="accountId", type=db.AttributeType.STRING),
            billing_mode=db.BillingMode.PROVISIONED,
            read_capacity=100,
            write_capacity=100,
            removal_policy=cdk.RemovalPolicy.DESTROY,
        )

        db.Table(
            self,
            "ConfigTable",
            table_name="consoleme_config_global",
            partition_key=db.Attribute(name="id", type=db.AttributeType.STRING),
            billing_mode=db.BillingMode.PROVISIONED,
            read_capacity=10,
            write_capacity=10,
            removal_policy=cdk.RemovalPolicy.DESTROY,
        )

        requests_table = db.Table(
            self,
            "RequestsTable",
            table_name="consoleme_policy_requests",
            partition_key=db.Attribute(name="request_id", type=db.AttributeType.STRING),
            sort_key=db.Attribute(name="arn", type=db.AttributeType.STRING),
            billing_mode=db.BillingMode.PROVISIONED,
            read_capacity=10,
            write_capacity=10,
            removal_policy=cdk.RemovalPolicy.DESTROY,
        )

        requests_table.add_global_secondary_index(
            index_name="arn-request_id-index",
            partition_key=db.Attribute(name="arn", type=db.AttributeType.STRING),
            read_capacity=123,
            write_capacity=123,
            projection_type=db.ProjectionType.ALL,
        )

        cache_table = db.Table(
            self,
            "CacheTable",
            table_name="consoleme_resource_cache",
            partition_key=db.Attribute(name="resourceId", type=db.AttributeType.STRING),
            sort_key=db.Attribute(name="resourceType", type=db.AttributeType.STRING),
            billing_mode=db.BillingMode.PROVISIONED,
            read_capacity=10,
            write_capacity=10,
            removal_policy=cdk.RemovalPolicy.DESTROY,
        )

        cache_table.add_global_secondary_index(
            index_name="arn-index",
            partition_key=db.Attribute(name="arn", type=db.AttributeType.STRING),
            read_capacity=123,
            write_capacity=123,
            projection_type=db.ProjectionType.ALL,
        )

        db.Table(
            self,
            "CloudTrailTable",
            table_name="consoleme_cloudtrail",
            partition_key=db.Attribute(name="arn", type=db.AttributeType.STRING),
            sort_key=db.Attribute(name="request_id", type=db.AttributeType.STRING),
            billing_mode=db.BillingMode.PROVISIONED,
            read_capacity=10,
            write_capacity=10,
            removal_policy=cdk.RemovalPolicy.DESTROY,
        )

        db.Table(
            self,
            "UsersTable",
            table_name="consoleme_users_global",
            partition_key=db.Attribute(name="username", type=db.AttributeType.STRING),
            billing_mode=db.BillingMode.PROVISIONED,
            read_capacity=5,
            write_capacity=5,
            removal_policy=cdk.RemovalPolicy.DESTROY,
        )



from typing import Dict

from consoleme.config import config
from consoleme.handlers.base import BaseAPIV1Handler
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.auth import (
    can_admin_policies,
    can_create_roles,
    can_delete_iam_principals,
    can_edit_dynamic_config,
)
from consoleme.lib.generic import get_random_security_logo, is_in_group
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.v2.user_profile import get_custom_page_header

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


class UserProfileHandler(BaseAPIV1Handler):
    async def get(self):
        """
        Provide information about site configuration for the frontend
        :return:
        """
        is_contractor = config.config_plugin().is_contractor(self.user)
        site_config = {
            "consoleme_logo": await get_random_security_logo(),
            "google_analytics": {
                "tracking_id": config.get("google_analytics.tracking_id"),
                "options": config.get("google_analytics.options", {}),
            },
            "documentation_url": config.get(
                "documentation_page", "https://hawkins.gitbook.io/consoleme/"
            ),
            "support_contact": config.get("support_contact"),
            "support_chat_url": config.get(
                "support_chat_url", "https://discord.com/invite/nQVpNGGkYu"
            ),
            "security_logo": config.get("security_logo.image"),
            "security_url": config.get("security_logo.url"),
            # If site_config.landing_url is set, users will be redirected to the landing URL after authenticating
            # on the frontend.
            "landing_url": config.get("site_config.landing_url"),
            "notifications": {
                "enabled": config.get("site_config.notifications.enabled"),
                "request_interval": config.get(
                    "site_config.notifications.request_interval", 60
                ),
            },
            "cloudtrail_denies_policy_generation": config.get(
                "celery.cache_cloudtrail_denies.enabled", False
            ),
        }

        custom_page_header: Dict[str, str] = await get_custom_page_header(
            self.user, self.groups
        )
        user_profile = {
            "site_config": site_config,
            "user": self.user,
            "can_logout": config.get("auth.set_auth_cookie", False),
            "is_contractor": is_contractor,
            "employee_photo_url": config.config_plugin().get_employee_photo_url(
                self.user
            ),
            "employee_info_url": config.config_plugin().get_employee_info_url(
                self.user
            ),
            "authorization": {
                "can_edit_policies": can_admin_policies(self.user, self.groups),
                "can_create_roles": can_create_roles(self.user, self.groups),
                "can_delete_iam_principals": can_delete_iam_principals(
                    self.user, self.groups
                ),
            },
            "pages": {
                "header": {
                    "custom_header_message_title": custom_page_header.get(
                        "custom_header_message_title", ""
                    ),
                    "custom_header_message_text": custom_page_header.get(
                        "custom_header_message_text", ""
                    ),
                    "custom_header_message_route": custom_page_header.get(
                        "custom_header_message_route", ""
                    ),
                },
                "groups": {
                    "enabled": config.get("headers.group_access.enabled", False)
                },
                "users": {"enabled": config.get("headers.group_access.enabled", False)},
                "policies": {
                    "enabled": config.get("headers.policies.enabled", True)
                    and not is_contractor
                },
                "self_service": {
                    "enabled": config.get("enable_self_service", True)
                    and not is_contractor
                },
                "api_health": {
                    "enabled": is_in_group(
                        self.user,
                        self.groups,
                        config.get("groups.can_edit_health_alert", []),
                    )
                },
                "audit": {
                    "enabled": is_in_group(
                        self.user, self.groups, config.get("groups.can_audit", [])
                    )
                },
                "config": {"enabled": can_edit_dynamic_config(self.user, self.groups)},
            },
            "accounts": await get_account_id_to_name_mapping(),
        }

        self.set_header("Content-Type", "application/json")
        self.write(user_profile)



from consoleme.config import config


async def authenticate_user_by_credentials(request):
    # If request is a POST and it has username, password, etc, then validate and return
    # If request doesn't have these parameters, render login page
    email = None
    groups = None

    if request.request.path == "/auth":
        # If request is post and has credentials, validate or return error
        pass

    if email and groups and config.get("auth.set_auth_cookie"):
        pass
        # encoded_cookie = await generate_jwt_token(email, groups)
        # request.set_cookie(config.get("auth_cookie_name", "consoleme_auth"), encoded_cookie)

    if request.request.path != "/auth":
        request.redirect("/auth")



"""
Compute stack for running ConsoleMe on ECS
"""

from aws_cdk import aws_applicationautoscaling as applicationautoscaling
from aws_cdk import aws_certificatemanager as acm
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_ecr_assets as ecr_assets
from aws_cdk import aws_ecs as ecs
from aws_cdk import aws_ecs_patterns as ecs_patterns
from aws_cdk import aws_elasticloadbalancingv2 as lb
from aws_cdk import aws_iam as iam
from aws_cdk import aws_logs as logs
from aws_cdk import core as cdk

from consoleme_ecs_cdk.service.constants import (
    DOCKER_IMAGE,
    MAX_CAPACITY,
    MIN_CAPACITY,
    USE_PUBLIC_DOCKER_IMAGE,
)


class ComputeStack(cdk.NestedStack):
    """
    Compute stack for running ConsoleMe on ECS
    """

    def __init__(
        self,
        scope: cdk.Construct,
        id: str,
        vpc: ec2.Vpc,
        s3_bucket_name: str,
        certificate: acm.Certificate,
        consoleme_alb: lb.ApplicationLoadBalancer,
        consoleme_sg: ec2.SecurityGroup,
        task_role_arn: str,
        task_execution_role_arn: str,
        **kwargs
    ) -> None:
        super().__init__(scope, id, **kwargs)

        # ECS Task definition and volumes

        if USE_PUBLIC_DOCKER_IMAGE is True:
            docker_image = ecs.ContainerImage.from_registry(DOCKER_IMAGE)
        else:
            docker_image = ecs.ContainerImage.from_docker_image_asset(
                ecr_assets.DockerImageAsset(
                    self, "ConsoleMeCustomImage", directory="../"
                )
            )

        imported_task_role = iam.Role.from_role_arn(
            self, "ImportedTaskRole", role_arn=task_role_arn
        )

        imported_task_execution_role = iam.Role.from_role_arn(
            self, "ImportedTaskExecutionRole", role_arn=task_execution_role_arn
        )

        consoleme_ecs_task_definition = ecs.FargateTaskDefinition(
            self,
            "ConsoleMeTaskDefinition",
            cpu=2048,
            memory_limit_mib=4096,
            execution_role=imported_task_execution_role,
            task_role=imported_task_role,
        )

        # ECS Container definition, service, target group and ALB attachment

        consoleme_ecs_task_definition.add_container(
            "Container",
            image=docker_image,
            privileged=False,
            port_mappings=[
                ecs.PortMapping(
                    container_port=8081, host_port=8081, protocol=ecs.Protocol.TCP
                )
            ],
            logging=ecs.LogDriver.aws_logs(
                stream_prefix="ContainerLogs-",
                log_retention=logs.RetentionDays.ONE_WEEK,
            ),
            environment={
                "SETUPTOOLS_USE_DISTUTILS": "stdlib",
                "CONSOLEME_CONFIG_S3": "s3://" + s3_bucket_name + "/config.yaml",
                "EC2_REGION": self.region,
            },
            working_directory="/apps/consoleme",
            command=[
                "bash",
                "-c",
                "python scripts/retrieve_or_decode_configuration.py; python consoleme/__main__.py",
            ],
        )

        consoleme_ecs_task_definition.add_container(
            "CeleryContainer",
            image=docker_image,
            privileged=False,
            logging=ecs.LogDriver.aws_logs(
                stream_prefix="CeleryContainerLogs-",
                log_retention=logs.RetentionDays.ONE_WEEK,
            ),
            environment={
                "SETUPTOOLS_USE_DISTUTILS": "stdlib",
                "CONSOLEME_CONFIG_S3": "s3://" + s3_bucket_name + "/config.yaml",
                "COLUMNS": "80",
                "EC2_REGION": self.region,
            },
            command=[
                "bash",
                "-c",
                "python scripts/retrieve_or_decode_configuration.py; python scripts/initialize_redis_oss.py; celery -A consoleme.celery_tasks.celery_tasks worker -l DEBUG -B -E --concurrency=8",
            ],
        )

        # ECS cluster

        cluster = ecs.Cluster(self, "Cluster", vpc=vpc)

        consoleme_imported_alb = (
            lb.ApplicationLoadBalancer.from_application_load_balancer_attributes(
                self,
                "ConsoleMeImportedALB",
                load_balancer_arn=consoleme_alb.load_balancer_arn,
                vpc=vpc,
                security_group_id=consoleme_sg.security_group_id,
                load_balancer_dns_name=consoleme_alb.load_balancer_dns_name,
            )
        )

        consoleme_ecs_service = ecs_patterns.ApplicationLoadBalancedFargateService(
            self,
            "Service",
            cluster=cluster,
            task_definition=consoleme_ecs_task_definition,
            load_balancer=consoleme_imported_alb,
            security_groups=[consoleme_sg],
            open_listener=False,
        )

        consoleme_ecs_service.target_group.configure_health_check(
            path="/", enabled=True, healthy_http_codes="200-302"
        )

        consoleme_ecs_service_scaling_target = applicationautoscaling.ScalableTarget(
            self,
            "AutoScalingGroup",
            max_capacity=MAX_CAPACITY,
            min_capacity=MIN_CAPACITY,
            resource_id="service/"
            + cluster.cluster_name
            + "/"
            + consoleme_ecs_service.service.service_name,
            scalable_dimension="ecs:service:DesiredCount",
            service_namespace=applicationautoscaling.ServiceNamespace.ECS,
            role=iam.Role(
                self,
                "AutoScaleRole",
                assumed_by=iam.ServicePrincipal(service="ecs-tasks.amazonaws.com"),
                description="Role for ECS auto scaling group",
                managed_policies=[
                    iam.ManagedPolicy.from_managed_policy_arn(
                        self,
                        "AutoScalingManagedPolicy",
                        managed_policy_arn="arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceAutoscaleRole",
                    )
                ],
            ),
        )

        applicationautoscaling.TargetTrackingScalingPolicy(
            self,
            "AutoScalingPolicy",
            scaling_target=consoleme_ecs_service_scaling_target,
            scale_in_cooldown=cdk.Duration.seconds(amount=10),
            scale_out_cooldown=cdk.Duration.seconds(amount=10),
            target_value=50,
            predefined_metric=applicationautoscaling.PredefinedMetric.ECS_SERVICE_AVERAGE_CPU_UTILIZATION,
        )

        consoleme_imported_alb.add_listener(
            "ConsoleMeALBListener",
            protocol=lb.ApplicationProtocol.HTTPS,
            port=443,
            certificates=[certificate],
            default_action=lb.ListenerAction.forward(
                target_groups=[consoleme_ecs_service.target_group]
            ),
        )



import json
import os
import time

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    WebpackBundleLookupError,
    WebpackError,
    WebpackLoaderBadStatsError,
    WebpackLoaderTimeoutError,
)

settings = {}

BASE_DIR = config.get(
    "webpack.loader.base_directory",
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
)
DEV_STATS = config.get("webpack.loader.dev_stats", "webpack-stats.dev.json")
PROD_STATS = config.get("webpack.loader.prod_stats", "webpack-stats.prod.json")
BUNDLE_DIR_NAME = config.get("webpack.loader.bundle_dir", "bundles/")


class WebpackLoader(object):
    _assets = {}

    def __init__(self, name, config):
        self.name = name
        self.config = config

    def load_assets(self):
        # TODO, get stats file path from config
        # TODO, select stats file based on development or production
        ENV_STAT_FILE = DEV_STATS if self.config.get("development") else PROD_STATS
        STATS_FILE = os.path.join(BASE_DIR, ENV_STAT_FILE)
        try:
            with open(STATS_FILE, encoding="utf-8") as f:
                return json.load(f)
        except IOError:
            raise IOError(
                f"Error reading {STATS_FILE}. Are you sure webpack has generated "
                "the file and the path is correct?"
            )

    def get_assets(self):
        # TODO, get assets from CACHE
        return self.load_assets()

    def filter_chunks(self, chunks):
        for chunk in chunks:
            chunk["url"] = self.get_chunk_url(chunk)
            yield chunk

    def get_chunk_url(self, chunk):
        public_path = chunk.get("publicPath")
        if public_path:
            return public_path

        relpath = "{0}{1}".format(BUNDLE_DIR_NAME, chunk["name"])
        # TODO: revisit this
        # return staticfiles_storage.url(relpath)
        return relpath

    def get_bundle(self, bundle_name):
        assets = self.get_assets()

        # poll when debugging and block request until bundle is compiled
        # or the build times out
        if self.config.get("development"):
            timeout = 0
            timed_out = False
            start = time.time()
            while assets["status"] == "compiling" and not timed_out:
                time.sleep(0.1)
                if timeout and (time.time() - timeout > start):
                    timed_out = True
                assets = self.get_assets()

            if timed_out:
                raise WebpackLoaderTimeoutError(
                    f"Timed Out. Bundle took more than {timeout} seconds " "to compile."
                )

        if assets.get("status") == "done":
            chunks = assets["chunks"].get(bundle_name, None)
            if chunks is None:
                raise WebpackBundleLookupError(
                    "Cannot resolve bundle {0}.".format(bundle_name)
                )
            return self.filter_chunks(chunks)

        elif assets.get("status") == "error":
            if "file" not in assets:
                assets["file"] = ""
            if "error" not in assets:
                assets["error"] = "Unknown Error"
            if "message" not in assets:
                assets["message"] = ""
            error = "{} in {} {}".format(
                assets["error"], assets["file"], assets["message"]
            )
            raise WebpackError(error)

        raise WebpackLoaderBadStatsError(
            "The stats file does not contain valid data. Make sure "
            "webpack-bundle-tracker plugin is enabled and try to run "
            "webpack again."
        )



"""
Setup tools for consoleme_ecs_service cdk library
"""

import setuptools

with open("../README.md") as fp:
    long_description = fp.read()


setuptools.setup(
    name="consoleme_ecs_cdk",
    version="0.0.1",
    description="AWS CDK stack for ConsoleMe ECS service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Avishay Bar",
    package_dir={"": "service"},
    packages=setuptools.find_packages(where="service"),
    install_requires=[
        "aws-cdk.core>=1.107.0",
        "aws-cdk.aws_s3>=1.107.0",
        "aws-cdk.aws-s3-deployment>=1.107.0",
        "aws-cdk.aws-iam>=1.107.0",
        "aws-cdk.aws-cognito>=1.107.0",
        "aws-cdk.aws_ec2>=1.107.0",
        "aws-cdk.aws_ecs>=1.107.0",
        "aws-cdk.aws-ecs-patterns>=1.107.0",
        "aws-cdk.aws-ecr-assets>=1.107.0",
        "aws-cdk.aws_efs>=1.107.0",
        "aws_cdk.aws_certificatemanager>=1.107.0",
        "aws_cdk.aws_route53>=1.107.0",
        "aws_cdk.aws_route53_targets>=1.107.0",
        "aws_cdk.aws_elasticloadbalancingv2>=1.107.0",
        "aws_cdk.aws_logs>=1.107.0",
        "aws_cdk.aws_kms>=1.107.0",
        "aws_cdk.aws_elasticache>=1.107.0",
        "aws_cdk.aws_dynamodb>=1.107.0",
        "aws_cdk.custom_resources>=1.107.0",
        "aws_cdk.aws_lambda>=1.107.0",
        "aws_cdk.aws-applicationautoscaling>=1.107.0",
        "PyYAML>=5.3.1",
        "pipreqs>=0.4.10",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)
